import re
from datetime import datetime
from io import BytesIO, StringIO
from typing import Generator, Optional

from anymail.inbound import AnymailInboundMessage
from django.conf import settings
from django.db.models import Model, Q
from dynamic_preferences.registries import global_preferences_registry
from slugify import slugify

from ..models import Source
from .abstract import AbstractDataBackend
from .utils import register


@register("Default Mail", save_data_in_import_source=True, passive_only=True)
class DataBackend(AbstractDataBackend):
    def get_files(
        self,
        execution_time: datetime,
        file_name_regex: Optional[str] = None,
        message: Optional[AnymailInboundMessage] = None,
        import_credential: Optional[Model] = None,
        import_email_as_file: Optional[bool] = False,
        **kwargs,
    ) -> Generator[tuple[str, BytesIO], None, None]:
        """

        Args:
            execution_time: The time at which this import was called
            file_name_regex: The regex applied on the imported file for validation. Defaults to False.
            message: The AnymailInboundMessage received from the anymail inbounc
            import_credential: Import credential attached to the calling source. Defaults to None.
            import_email_as_file: If false, import the attachment as import source file. Otherwise, import the whole email as file
            **kwargs:

        Returns:

        """
        if message:
            if import_email_as_file:
                filename = f"{slugify(message.subject)}_{datetime.timestamp(execution_time)}.eml"
                yield filename, StringIO(message.as_string())
            elif file_name_regex:
                attachments = message.attachments
                attachments.extend(message.inline_attachments.values())

                for attachment in attachments:
                    f = attachment.as_uploaded_file()
                    f_name = attachment.get_filename()
                    result = re.findall(file_name_regex, f_name)
                    if len(result) > 0:
                        yield f_name, f.file


def is_sender_allowed(from_email: str, whitelisted_emails: list[str], admin_mails: list[str]) -> bool:
    if from_email in admin_mails:
        return True
    for whitelisted_email in whitelisted_emails:
        if re.search(whitelisted_email, from_email):
            return True
    return False


def handle_inbound(sender, event, esp_name, **kwargs):
    spam_detected = False
    if (message := event.message) and (from_email := message.from_email.addr_spec) and (subject := message.subject):
        if spam_score := getattr(settings, "WBIMPORT_EXPORT_MAILBACKEND_SPAMSCORE", None):
            spam_detected = message.spam_detected or message.spam_score >= spam_score
        if not spam_detected:
            admin_emails = global_preferences_registry.manager()["io__administrator_mails"].split(";")
            conditions = Q(import_parameters__inbound_address__isnull=True)
            for t in message.to:
                conditions |= Q(import_parameters__inbound_address__contains=t.addr_spec)

            sources = Source.objects.filter(
                conditions & Q(data_backend__backend_class_path="wbcore.contrib.io.backends.mail") & Q(is_active=True)
            )
            if s := re.search(r"\[([0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12})\]", subject):
                sources = sources.filter(uuid=s.group(1))

            for source in sources:
                if is_sender_allowed(from_email, source.import_parameters.get("whitelisted_emails", []), admin_emails):
                    source.trigger_workflow(message=message)



from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import StringPreference

io = Section("io")


@global_preferences_registry.register
class AdministratorMailsPreference(StringPreference):
    section = io
    name = "administrator_mails"
    default = ""

    verbose_name = "Administrator mails"
    help_text = "The mails (as comma separated string) of the person allow to send directly to the mail backend"



# Generated by Django 4.1.9 on 2023-06-02 06:15

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0006_calendaritem_is_deletable"),
    ]

    operations = [
        migrations.RemoveField(name="endpoint_basename", model_name="calendaritem"),
        migrations.AlterModelOptions(
            name="calendaritem",
            options={
                "permissions": [("administrate_confidential_items", "Can administrate calendar items")],
                "verbose_name": "Calendar Item",
                "verbose_name_plural": "Calendar Items",
            },
        ),
    ]



from contextlib import suppress
from typing import TYPE_CHECKING, Generic, Iterator, Type, TypeVar

from django.utils.module_loading import import_string

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from wbcore.contrib.dataloader.dataloaders import Dataloader
    from wbcore.contrib.dataloader.models import Entity


T = TypeVar("T")


class DataloaderProxy(Generic[T]):
    """Manages and facilitates data loading from remote resources using dataloaders.

    This class serves as a proxy for various dataloader instances. Each dataloader is
    a subclass of `Dataloader`, designed to fetch data from a remote resource. The
    `DataloaderProxy` class enables the splicing of a queryset into parts, where each
    part uses a specific dataloader, thus abstracting the complexity of data fetching
    from multiple sources.

    The key responsibility of this class includes:
    - Splicing a queryset into segments, ensuring each segment uses the corresponding dataloader.
    - Providing an external interface to interact with different dataloaders, which allows
      for efficient data retrieval and processing.

    To utilize this class, it should be subclassed and associated with a Django model
    that can be processed by the `Dataloader` subclasses. This setup facilitates the
    dynamic and efficient retrieval of data, tailored to the specific needs of the application.

    Attributes:
        queryset (QuerySet[Entity]): A Django queryset that represents a collection of
            entities to be processed by the dataloaders.
    """

    def __init__(self, queryset: "QuerySet[Entity]"):
        self.queryset = queryset

    @staticmethod
    def load_dataloader(path: str) -> Type["Dataloader"]:
        """Loads a Dataloader object from a given path.

        This static method imports a Dataloader class from the specified path string.

        Args:
            path (str): The dot-delimited path to the Dataloader class to be imported.
                For example, 'module.submodule.ClassName'.

        Returns:
            Type["Dataloader"]: An instance of the Dataloader class loaded from the given path.

        Raises:
            ImportError: If the path does not correspond to a valid class.
        """
        return import_string(path)

    def iterate_dataloaders(self, dl_name: str) -> Iterator[T]:
        """Iterates over distinct dataloaders specified by a given name.

        This method retrieves a distinct list of paths associated with the specified
        dataloader name, then iterates over these paths. For each path, it loads a
        dataloader instance and filters the queryset based on the path, then yields
        this dataloader.

        The type `T` represents the specific type of the dataloaders that will be loaded and
        iterated over.

        Args:
            dl_name (str): The name of the dataloader to be iterated over. This name
                is used to construct the path key for querying distinct dataloader paths.

        Returns:
            Iterator[T]: An iterator of dataloader instances of the generic type `T` corresponding
            to each distinct path found in the queryset.
        """
        dl_path = f"dl_parameters__{dl_name}__path"
        for path in self.queryset.distinct(dl_path).values_list(dl_path, flat=True):
            with suppress(AttributeError):
                yield self.load_dataloader(path)(self.queryset.filter(**{dl_path: path}))



# Generated by Django 4.2.10 on 2024-03-06 07:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("documents", "0002_documentmodelrelationship_primary_and_more"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="documentmodelrelationship",
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name="documentmodelrelationship",
            constraint=models.UniqueConstraint(
                condition=models.Q(("primary", True)),
                fields=("content_type", "object_id"),
                name="unique_primary_relationship",
            ),
        ),
        migrations.AddConstraint(
            model_name="documentmodelrelationship",
            constraint=models.UniqueConstraint(
                fields=("document", "content_type", "object_id"), name="unique_relationship"
            ),
        ),
    ]



from datetime import timedelta

import pytest
from faker import Faker
from wbcore.contrib.authentication.tasks import delete_unregistered_user_account

from ..models import User

fake = Faker()


@pytest.mark.django_db
def test_delete_unregistered_user_account(user_factory):
    val_datetime = fake.date_time()

    registered_user = user_factory.create(is_register=True, is_active=True)
    unregistered_but_active_user = user_factory.create(
        is_register=False, is_active=True
    )  # in case we have (but we shouldn't) unregistered but active user, we don't clean them
    unregistered_user_in_retention_period = user_factory.create(
        is_register=False, is_active=False, date_joined=val_datetime
    )  # unregistered user created within the retention period
    unregistered_user_obselete = user_factory.create(
        is_register=False, is_active=False, date_joined=val_datetime - timedelta(seconds=1)
    )  # unregistered user created outside of the retention period
    delete_unregistered_user_account(prune_user_account_before_datetime=val_datetime)
    assert set(User.objects.exclude(email="AnonymousUser")) == {
        registered_user,
        unregistered_but_active_user,
        unregistered_user_in_retention_period,
    }
    with pytest.raises(User.DoesNotExist):
        unregistered_user_obselete.refresh_from_db()



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.workflow.models import ProcessStep
from wbcore.contrib.workflow.viewsets.display.process import (
    get_state_formatting,
    get_state_legend,
)


@pytest.mark.django_db
class TestProcessStep:
    api_factory = APIRequestFactory()

    def test_get_state_legend(self):
        color_map = [
            (ProcessStep.StepState.FAILED, WBColor.GREEN_LIGHT.value),
            (ProcessStep.StepState.CANCELED, WBColor.BLUE_LIGHT.value),
        ]
        assert set((y.icon, y.label, y.value) for y in get_state_legend(color_map)[0].items) == {
            (WBColor.GREEN_LIGHT.value, ProcessStep.StepState.FAILED.label, ProcessStep.StepState.FAILED.value),
            (WBColor.BLUE_LIGHT.value, ProcessStep.StepState.CANCELED.label, ProcessStep.StepState.CANCELED.value),
        }

    def test_get_state_formatting(self):
        color_map = [
            (ProcessStep.StepState.WAITING, WBColor.GREY.value),
            (ProcessStep.StepState.ACTIVE, WBColor.YELLOW_DARK.value),
        ]
        assert set(
            (y.style["backgroundColor"], y.condition[1]) for y in get_state_formatting(color_map)[0].formatting_rules
        ) == {
            (WBColor.GREY.value, ProcessStep.StepState.WAITING.value),
            (WBColor.YELLOW_DARK.value, ProcessStep.StepState.ACTIVE.value),
        }

    # @patch("wbcore.contrib.workflow.viewsets.display.process.split_list_into_grid_template_area_sublists")
    # def test_process_step_display_instance_injection(
    #     self, mock_split, data_factory, user_step_factory, process_step_factory
    # ):
    #     request = self.api_factory.get("")
    #     request.user = get_or_create_superuser()
    #     request.query_params = {}
    #     step = user_step_factory()
    #     process_step = process_step_factory(step=step)
    #     data_factory(workflow=process_step.step.workflow)
    #     mvs = ProcessStepModelViewSet(request=request, kwargs={"pk": process_step.pk})
    #     display = mvs.display_config_class(mvs, request).get_instance_display()
    #     assert display.pages[0].layouts["default"].grid_template_areas == step.display.grid_template_areas
    #     assert not mock_split.called

    # @patch("wbcore.contrib.workflow.viewsets.display.process.split_list_into_grid_template_area_sublists")
    # def test_process_step_display_no_userstep(self, mock_split, random_child_step_factory, process_step_factory):
    #     request = self.api_factory.get("")
    #     request.user = get_or_create_superuser()
    #     request.query_params = {}
    #     step = random_child_step_factory(exclude_factories=[UserStepFactory])
    #     process_step = process_step_factory(step=step)
    #     mvs = ProcessStepModelViewSet(request=request, kwargs={"pk": process_step.pk})
    #     display = mvs.display_config_class(mvs, request).get_instance_display()
    #     display_fields = set(itertools.chain.from_iterable(display.pages[0].layouts["default"].grid_template_areas))
    #     person_fields = set(PersonModelSerializer.Meta.fields)
    #     assert not mock_split.called
    #     assert display_fields.intersection(person_fields) == {"id"}

    # @patch("wbcore.contrib.workflow.viewsets.display.process.split_list_into_grid_template_area_sublists")
    # def test_process_step_display_no_display(self, mock_split, user_step_factory, process_step_factory):
    #     request = self.api_factory.get("")
    #     request.user = get_or_create_superuser()
    #     request.query_params = {}
    #     step = user_step_factory(display=None)
    #     process_step = process_step_factory(step=step)
    #     mvs = ProcessStepModelViewSet(request=request, kwargs={"pk": process_step.pk})
    #     display = mvs.display_config_class(mvs, request).get_instance_display()
    #     display_fields = set(itertools.chain.from_iterable(display.pages[0].layouts["default"].grid_template_areas))
    #     person_fields = set(PersonModelSerializer.Meta.fields)
    #     assert not mock_split.called
    #     assert display_fields.intersection(person_fields) == {"id"}

    # @patch("wbcore.contrib.workflow.viewsets.display.process.split_list_into_grid_template_area_sublists")
    # def test_process_step_display_data_injection(
    #     self, mock_split, data_factory, user_step_factory, process_step_factory
    # ):
    #     request = self.api_factory.get("")
    #     request.user = get_or_create_superuser()
    #     request.query_params = {}
    #     step = user_step_factory(display=None)
    #     process_step = process_step_factory(step=step)
    #     data1 = data_factory(workflow=process_step.process.workflow)
    #     data2 = data_factory(workflow=process_step.process.workflow)
    #     data_factory()
    #     data_sublist = ["a", "b", "c"]
    #     mock_split.return_value = data_sublist
    #     mvs = ProcessStepModelViewSet(request=request, kwargs={"pk": process_step.pk})
    #     display = mvs.display_config_class(mvs, request).get_instance_display()
    #     display_fields = set(itertools.chain.from_iterable(display.pages[0].layouts["default"].grid_template_areas))
    #     person_fields = set(PersonModelSerializer.Meta.fields)
    #     assert display_fields.intersection(person_fields) == {"id"}
    #     assert mock_split.call_args.args[0] == [f"data__{str(data1.pk)}", f"data__{str(data2.pk)}"]
    #     for key in data_sublist:
    #         assert key in display_fields

    # @patch("wbcore.contrib.workflow.viewsets.display.process.split_list_into_grid_template_area_sublists")
    # def test_process_step_display_no_display_no_data(
    #     self, mock_split, data_factory, user_step_factory, process_step_factory
    # ):
    #     request = self.api_factory.get("")
    #     request.user = get_or_create_superuser()
    #     request.query_params = {}
    #     step = user_step_factory(display=None)
    #     process_step = process_step_factory(step=step)
    #     data_factory()
    #     data_factory()
    #     mvs = ProcessStepModelViewSet(request=request, kwargs={"pk": process_step.pk})
    #     display = mvs.display_config_class(mvs, request).get_instance_display()
    #     display_fields = set(itertools.chain.from_iterable(display.pages[0].layouts["default"].grid_template_areas))
    #     person_fields = set(PersonModelSerializer.Meta.fields)
    #     assert display_fields.intersection(person_fields) == {"id"}
    #     assert not mock_split.called



from unittest.mock import patch

import pytest
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.directory.factories import ClientFactory
from wbcore.contrib.workflow.models import ProcessStep, Step
from wbcore.contrib.workflow.workflows import (
    manager_of_instance_assignee,
    random_group_member,
    weighted_random,
)


@pytest.mark.django_db
class TestWorkflowAssignees:
    def test_manager_of_instance_assignee_person(self, process_step_factory):
        kwargs = {"assignee_field": "profile", "assignee_type": "entry"}
        attached_instance = UserFactory()
        process_step = process_step_factory(process__instance=attached_instance)
        manager_list = ClientFactory.create_batch(3, clients=[attached_instance.profile])
        UserFactory(profile=manager_list[1], is_active=False)
        UserFactory(profile=manager_list[2])
        assert manager_of_instance_assignee(process_step, **kwargs) == manager_list[2].user_account

    def test_manager_of_instance_assignee_no_kwargs(self, process_step_factory):
        process_step = process_step_factory()
        assert not manager_of_instance_assignee(process_step)
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_manager_of_instance_assignee_no_assignee_field(self, process_step_factory):
        kwargs = {"test": "profile", "assignee_type": "entry"}
        process_step = process_step_factory()
        assert not manager_of_instance_assignee(process_step, **kwargs)
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_manager_of_instance_assignee_no_assignee_type(self, process_step_factory):
        kwargs = {"assignee_field": "profile", "test": "entry"}
        process_step = process_step_factory()
        assert not manager_of_instance_assignee(process_step, **kwargs)
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_manager_of_instance_assignee_wrong_assignee_field(self, process_step_factory):
        kwargs = {"assignee_field": "profile", "assignee_type": "entry"}
        process_step = process_step_factory()
        assert not manager_of_instance_assignee(process_step, **kwargs)
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_manager_of_instance_assignee_wrong_assignee_type(self, process_step_factory):
        kwargs = {"assignee_field": "profile", "assignee_type": "test"}
        attached_instance = UserFactory()
        process_step = process_step_factory(process__instance=attached_instance)
        assert not manager_of_instance_assignee(process_step, **kwargs)
        assert process_step.state == ProcessStep.StepState.FAILED

    @pytest.mark.parametrize(
        "bad_states", [ProcessStep.StepState.FAILED, ProcessStep.StepState.CANCELED, ProcessStep.StepState.WAITING]
    )
    @pytest.mark.parametrize(
        "bad_types",
        [
            Step.StepType.DECISIONSTEP,
            Step.StepType.EMAILSTEP,
            Step.StepType.FINISHSTEP,
            Step.StepType.JOINSTEP,
            Step.StepType.SPLITSTEP,
            Step.StepType.SCRIPTSTEP,
        ],
    )
    @patch("wbcore.contrib.workflow.workflows.assignees.choices")
    def test_weighted_random(self, mock_choices, user_step_factory, process_step_factory, bad_states, bad_types):
        group = GroupFactory()
        process_step = process_step_factory(group=group)
        assignee1 = UserFactory(groups=[group])
        assignee2 = UserFactory(groups=[group])
        user_step = user_step_factory()
        process_step_factory(
            state=ProcessStep.StepState.ACTIVE,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
            assignee=assignee1,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=bad_states,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.ACTIVE,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.ACTIVE,
            process=process_step.process,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step__step_type=bad_types,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
        )
        assert weighted_random(process_step)
        assert len(mock_choices.call_args.args) == 1
        assert len(mock_choices.call_args.kwargs) == 1
        assignees_with_weights = set(zip(mock_choices.call_args.args[0], mock_choices.call_args.kwargs["weights"]))
        assert assignees_with_weights == {(assignee1, 2 / 3), (assignee2, 1 / 3)}

    @pytest.mark.parametrize(
        "bad_states", [ProcessStep.StepState.FAILED, ProcessStep.StepState.CANCELED, ProcessStep.StepState.WAITING]
    )
    @pytest.mark.parametrize(
        "bad_types",
        [
            Step.StepType.DECISIONSTEP,
            Step.StepType.EMAILSTEP,
            Step.StepType.FINISHSTEP,
            Step.StepType.JOINSTEP,
            Step.StepType.SPLITSTEP,
            Step.StepType.SCRIPTSTEP,
        ],
    )
    @patch("wbcore.contrib.workflow.workflows.assignees.choices")
    def test_weighted_random_no_past_ocurrences(
        self, mock_choices, user_step_factory, process_step_factory, bad_states, bad_types
    ):
        group = GroupFactory()
        process_step = process_step_factory(group=group)
        assignee1 = UserFactory(groups=[group])
        assignee2 = UserFactory(groups=[group])
        user_step = user_step_factory()

        process_step_factory(
            state=bad_states,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.ACTIVE,
            group=process_step.group,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.ACTIVE,
            process=process_step.process,
            step=user_step,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step__step_type=bad_types,
            assignee=assignee2,
        )
        process_step_factory(
            state=ProcessStep.StepState.FINISHED,
            process=process_step.process,
            group=process_step.group,
            step=user_step,
        )
        assert weighted_random(process_step)
        assert len(mock_choices.call_args.args) == 1
        assert set(mock_choices.call_args.args[0]) == {assignee1, assignee2}
        assert not mock_choices.call_args.kwargs

    @patch("wbcore.contrib.workflow.workflows.assignees.choices")
    def test_weighted_random_no_group(self, mock_choices, process_step_factory):
        process_step = process_step_factory()
        assert not weighted_random(process_step)
        assert not mock_choices.called
        assert process_step.state == ProcessStep.StepState.FAILED

    @patch("wbcore.contrib.workflow.workflows.assignees.choices")
    def test_weighted_random_no_group_members(self, mock_choices, process_step_factory):
        process_step = process_step_factory(group=GroupFactory())
        assert not weighted_random(process_step)
        assert not mock_choices.called
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_random_group_member(self, process_step_factory):
        group = GroupFactory()
        process_step = process_step_factory(group=group)
        UserFactory(groups=[group])
        UserFactory(groups=[group])
        assert random_group_member(process_step)

    def test_random_group_member_no_group(self, process_step_factory):
        process_step = process_step_factory()
        assert not random_group_member(process_step)
        assert process_step.state == ProcessStep.StepState.FAILED

    def test_random_group_member_no_group_members(self, process_step_factory):
        process_step = process_step_factory(group=GroupFactory())
        assert not random_group_member(process_step)
        assert process_step.state == ProcessStep.StepState.FAILED



from django.db.models import Avg, F, Max, Min, QuerySet, Sum
from wbcore import viewsets
from wbcore.contrib.example_app.filters import (
    PlayerFilter,
    PlayerTeamFilter,
    SportPersonFilter,
    SportPersonRepresentationFilter,
)
from wbcore.contrib.example_app.models import Player, SportPerson, Team
from wbcore.contrib.example_app.serializers import (
    PlayerModelSerializer,
    PlayerRepresentationSerializer,
    SportPersonModelSerializer,
    SportPersonRepresentationSerializer,
    SportPersonTooltipSerializer,
    TreeViewPlayerModelSerializer,
)
from wbcore.contrib.example_app.viewsets.buttons import PlayerButtonConfig
from wbcore.contrib.example_app.viewsets.displays import (
    PlayerDisplayConfig,
    PlayerTeamDisplayConfig,
    SportPersonDisplayConfig,
    SportPersonToolTipDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import PlayerTeamEndpointConfig
from wbcore.contrib.example_app.viewsets.titles import (
    PlayerTeamTitleConfig,
    PlayerTitleConfig,
    SportPersonTitleConfig,
)
from wbcore.utils.strings import format_number
from wbcore.viewsets.mixins import OrderableMixin, ReparentMixin


class SportPersonModelViewSet(viewsets.ModelViewSet):
    display_config_class = SportPersonDisplayConfig
    queryset = SportPerson.objects.all()
    search_fields = ("first_name", "last_name")
    serializer_class = SportPersonModelSerializer
    title_config_class = SportPersonTitleConfig
    ordering_fields = (
        "last_name",
        "first_name",
        "roles__title",
    )
    ordering = ("last_name", "first_name")
    filterset_class = SportPersonFilter

    def get_queryset(self) -> QuerySet[SportPerson]:
        return super().get_queryset().prefetch_related("roles")


class SportPersonRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = SportPerson.objects.all()
    serializer_class = SportPersonRepresentationSerializer
    search_fields = ("computed_str",)
    filterset_class = SportPersonRepresentationFilter


class SportPersonToolTipViewset(viewsets.ModelViewSet):
    IDENTIFIER = "example_app:persontooltip"
    queryset = SportPerson.objects.all()
    display_config_class = SportPersonToolTipDisplayConfig
    serializer_class = SportPersonTooltipSerializer
    title_config_class = SportPersonTitleConfig


class PlayerModelViewSet(SportPersonModelViewSet):
    display_config_class = PlayerDisplayConfig
    queryset = Player.objects.all()
    search_fields = SportPersonModelViewSet.search_fields + ("current_team__name",)
    serializer_class = PlayerModelSerializer
    title_config_class = PlayerTitleConfig
    ordering_fields = SportPersonModelViewSet.ordering_fields + (
        "position",
        "current_team__name",
        "former_teams__name",
        "transfer_value",
    )
    ordering = ("last_name", "first_name", "current_team__name", "id")
    filterset_class = PlayerFilter
    button_config_class = PlayerButtonConfig

    def get_aggregates(self, queryset: QuerySet[Player], paginated_queryset: QuerySet[Player]):
        transfer_sum = queryset.aggregate(s=Sum(F("transfer_value")))["s"] or 0
        transfer_average = queryset.aggregate(a=Avg(F("transfer_value")))["a"] or 0
        transfer_max = queryset.aggregate(m=Max(F("transfer_value")))["m"] or 0
        transfer_min = queryset.aggregate(m=Min(F("transfer_value")))["m"] or 0
        return {
            "transfer_value": {
                "Σ": format_number(transfer_sum),
                "⌀": format_number(transfer_average),
                "Max": format_number(transfer_max),
                "Min": format_number(transfer_min),
            }
        }

    def get_queryset(self) -> QuerySet[Player]:
        return super().get_queryset().select_related("current_team").prefetch_related("former_teams")


class PlayerRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Player.objects.all()
    serializer_class = PlayerRepresentationSerializer
    search_fields = ("computed_str",)


class PlayerTeamModelViewSet(PlayerModelViewSet):
    display_config_class = PlayerTeamDisplayConfig
    ordering_fields = SportPersonModelViewSet.ordering_fields + (
        "position",
        "former_teams__name",
        "transfer_value",
    )
    ordering = ("last_name", "first_name", "id")
    filterset_class = PlayerTeamFilter
    title_config_class = PlayerTeamTitleConfig
    endpoint_config_class = PlayerTeamEndpointConfig

    def get_queryset(self) -> QuerySet[Player]:
        queryset = super().get_queryset()
        if team_id := self.kwargs.get("team_id"):
            queryset = queryset.filter(current_team=team_id)
        return queryset


class TreeViewPlayerModelViewSet(OrderableMixin, ReparentMixin, viewsets.ModelViewSet):
    PARENT_FIELD = "current_team"
    PARENT_MODEL = Team
    ordering = ("order", "computed_str", "id")
    filterset_class = PlayerFilter
    queryset = Player.objects.all()
    serializer_class = TreeViewPlayerModelSerializer



from reversion.models import Revision, Version
from wbcore import filters as wb_filters


class RevisionFilterSet(wb_filters.FilterSet):
    user = wb_filters.ModelChoiceFilter(
        queryset=Revision.objects.all(),
        endpoint="wbcore:revisionrepresentation-list",
        value_key="id",
        label_key="{{user}} - {{date_created}}",
    )
    profile = wb_filters.NumberFilter(method="filter_profile")

    def filter_profile(self, queryset, name, value):
        return queryset.filter(user__profile=value)

    class Meta:
        model = Revision
        fields = {"date_created": ["gte", "exact", "lte"]}


class VersionFilterSet(wb_filters.FilterSet):
    content_type = wb_filters.NumberFilter(method="filter_content_type")
    exclude_id = wb_filters.NumberFilter(method="filter_exclude_id")

    def filter_content_type(self, queryset, name, value):
        if value:
            return queryset.filter(content_type=value)
        return queryset

    def filter_exclude_id(self, queryset, name, value):
        if value:
            return queryset.exclude(id=value)
        return queryset

    class Meta:
        model = Version
        fields = {"content_type": ["exact"], "object_id": ["exact"]}



from django.contrib import admin
from wbcore.contrib.workflow.models import Process, ProcessStep


@admin.register(Process)
class ProcessAdmin(admin.ModelAdmin):
    search_fields = ("id", "workflow")
    list_display = (
        "id",
        "workflow",
        "state",
        "started",
        "finished",
        "instance_id",
        "content_type",
        "instance",
        "preserved_instance",
    )


@admin.register(ProcessStep)
class ProcessStepAdmin(admin.ModelAdmin):
    search_fields = ("id", "process", "step")
    list_display = (
        "id",
        "process",
        "step",
        "state",
        "error_message",
        "started",
        "finished",
        "assignee",
        "group",
        "permission",
        "status",
    )



# Generated by Django 4.1.8 on 2023-04-06 12:07

import django.db.models.deletion
import wbcore.contrib.currency.import_export.handlers.currency
import wbcore.contrib.currency.import_export.handlers.currency_fx_rates
import wbcore.utils.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("io", "0001_initial_squashed"),
    ]

    operations = [
        migrations.CreateModel(
            name="Currency",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("title", models.CharField(max_length=255)),
                ("symbol", models.CharField(blank=True, max_length=10, null=True)),
                ("key", models.CharField(max_length=3, unique=True)),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="io.importsource",
                    ),
                ),
            ],
            options={
                "verbose_name": "Currency",
                "verbose_name_plural": "Currencies",
                "ordering": ("title",),
            },
            bases=(
                wbcore.utils.models.LabelKeyMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="CurrencyFXRates",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("date", models.DateField()),
                ("value", models.DecimalField(decimal_places=6, max_digits=20)),
                (
                    "currency",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fx_rates",
                        to="currency.currency",
                    ),
                ),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="io.importsource",
                    ),
                ),
            ],
            options={
                "verbose_name": "FX Rate",
                "verbose_name_plural": "FX Rates",
            },
            bases=(models.Model,),
        ),
        migrations.AddIndex(
            model_name="currencyfxrates",
            index=models.Index(fields=["date", "currency"], name="currency_fx_rate_idx"),
        ),
        migrations.AlterUniqueTogether(
            name="currencyfxrates",
            unique_together={("date", "currency")},
        ),
    ]



from django.conf import settings
from django.contrib.staticfiles import finders
from django.db.utils import IntegrityError
from django.http.response import HttpResponse
from django.utils.module_loading import import_string
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from wbcore.contrib.notifications.models.tokens import NotificationUserToken


class NotificationTokenAPIView(APIView):
    """An APIView that returns the configuration needed for a frontend to generate a token and to upload said token"""

    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(import_string(settings.NOTIFICATION_BACKEND).get_configuration())

    def post(self, request: Request) -> Response:
        try:
            token: str = request.data["token"]  # type: ignore

            existing_instance = NotificationUserToken.objects.filter(
                user=request.user,
                token__startswith=token.split(":")[0],
            )

            if existing_instance.exists():
                existing_instance.update(token=token)
            else:
                NotificationUserToken.objects.create(
                    user=request.user,
                    token=token,
                    device_type=request.data["device_type"],  # type: ignore
                )
            return Response({}, status=status.HTTP_201_CREATED)
        except IntegrityError:
            return Response({}, status=status.HTTP_200_OK)
        except KeyError:
            return Response({}, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request: Request) -> Response:
        queryset = NotificationUserToken.objects.filter(user=request.user)

        if device_type := request.GET.get("device_type", None):  # type: ignore
            queryset = queryset.filter(device_type=device_type)

        queryset.delete()
        return Response({}, status=status.HTTP_204_NO_CONTENT)


def service_worker_view(request, service_worker_name: str | None = None):
    file_path = finders.find(f"notifications/{service_worker_name}")
    if file_path:
        with open(file_path) as f:
            response = HttpResponse(content=f.read(), content_type="application/javascript")
            response.headers["Service-Worker-Allowed"] = "../" * ((request.get_full_path()).count("/") - 1)
            return response
    return HttpResponse(status=status.HTTP_400_BAD_REQUEST)



from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.workflow.models import Condition, Step, Transition


class TransitionFilter(wb_filters.FilterSet):
    from_step = wb_filters.ModelMultipleChoiceFilter(
        label=_("From"),
        queryset=Step.objects.all(),
        endpoint=Step.get_representation_endpoint(),
        value_key=Step.get_representation_value_key(),
        label_key=Step.get_representation_label_key(),
    )
    to_step = wb_filters.ModelMultipleChoiceFilter(
        label=_("To"),
        queryset=Step.objects.all(),
        endpoint=Step.get_representation_endpoint(),
        value_key=Step.get_representation_value_key(),
        label_key=Step.get_representation_label_key(),
    )
    condition = wb_filters.ModelChoiceFilter(
        label=_("Associated Condition"),
        queryset=Condition.objects.all(),
        endpoint=Condition.get_representation_endpoint(),
        value_key=Condition.get_representation_value_key(),
        label_key=Condition.get_representation_label_key(),
        method="filter_by_condition",
    )

    def filter_by_condition(self, queryset: QuerySet[Transition], name, value: Condition) -> QuerySet[Transition]:
        if value:
            return queryset.filter(associated_conditions=value)
        return queryset

    class Meta:
        model = Transition
        fields = {
            "name": ["exact", "icontains"],
            "icon": ["icontains"],
        }



from unittest.mock import patch

import pytest
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.models import Process, ProcessStep
from wbcore.contrib.workflow.sites import workflow_site


@pytest.mark.django_db
class TestFinishStep:
    @pytest.mark.parametrize(
        "finished_state",
        [ProcessStep.StepState.FAILED, ProcessStep.StepState.CANCELED, ProcessStep.StepState.FINISHED],
    )
    @pytest.mark.parametrize("unfinished_state", [ProcessStep.StepState.ACTIVE, ProcessStep.StepState.WAITING])
    @patch("wbcore.contrib.workflow.models.step.Step.set_finished")
    def test_run(self, mock_finish, process_step_factory, finish_step_factory, unfinished_state, finished_state):
        step = finish_step_factory()
        process_step = process_step_factory(step=step)
        process_step_factory(process=process_step.process, state=finished_state)
        process_step_factory(step=step, state=unfinished_state)
        step.run(process_step)
        assert mock_finish.call_args.args == (process_step,)

    @pytest.mark.parametrize("unfinished_state", [ProcessStep.StepState.ACTIVE, ProcessStep.StepState.WAITING])
    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_run_failed(self, mock_failed, process_step_factory, finish_step_factory, unfinished_state):
        step = finish_step_factory()
        process_step = process_step_factory(step=step)
        process_step_factory(process=process_step.process, state=unfinished_state)
        step.run(process_step)
        assert mock_failed.call_args.args[0] == process_step

    @pytest.mark.parametrize(
        "unfinished_state",
        [
            ProcessStep.StepState.ACTIVE,
            ProcessStep.StepState.WAITING,
            ProcessStep.StepState.CANCELED,
            ProcessStep.StepState.FAILED,
        ],
    )
    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_finish_not_finished(self, mock_finish, unfinished_state, process_step_factory, finish_step_factory):
        step = finish_step_factory(write_preserved_instance=True)
        process_step = process_step_factory(
            process__preserved_instance={"first_name": "Test"},
            step=step,
            process__finished=None,
            state=unfinished_state,
            process__state=Process.ProcessState.ACTIVE,
        )
        step.finish(process_step)
        assert mock_finish.call_args.args == (process_step,)
        assert process_step.process.finished is None
        assert process_step.process.instance.first_name != "Test"
        assert process_step.process.state == Process.ProcessState.ACTIVE

    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_finish_write_preserved_instance(self, mock_finish, process_step_factory, finish_step_factory):
        step = finish_step_factory(write_preserved_instance=True, workflow__preserve_instance=True)
        process_step = process_step_factory(
            step=step,
            process__finished=None,
            state=ProcessStep.StepState.FINISHED,
            process__workflow=step.workflow,
        )
        old_name = process_step.process.instance.first_name
        workflow_site.registered_model_classes_serializer_map[
            Person
        ] = "wbcore.contrib.directory.serializers.PersonModelSerializer"
        step.finish(process_step)
        assert mock_finish.call_args.args == (process_step,)
        assert process_step.process.finished is not None
        assert process_step.process.instance.first_name != old_name
        assert process_step.process.instance.first_name == process_step.process.preserved_instance["first_name"]
        assert process_step.process.state == Process.ProcessState.FINISHED

    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_finish_do_not_write(self, mock_finish, process_step_factory, finish_step_factory):
        step = finish_step_factory(write_preserved_instance=False, workflow__preserve_instance=True)
        process_step = process_step_factory(
            step=step,
            process__finished=None,
            state=ProcessStep.StepState.FINISHED,
        )
        old_name = process_step.process.instance.first_name
        step.finish(process_step)
        assert mock_finish.call_args.args == (process_step,)
        assert process_step.process.finished is not None
        assert process_step.process.instance.first_name == old_name
        assert process_step.process.state == Process.ProcessState.FINISHED

    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_finish_no_preserved_instance(self, mock_finish, process_step_factory, finish_step_factory):
        step = finish_step_factory(write_preserved_instance=True, workflow__preserve_instance=True)
        process_step = process_step_factory(
            step=step, process__finished=None, state=ProcessStep.StepState.FINISHED, process__preserved_instance=None
        )
        old_name = process_step.process.instance.first_name
        step.finish(process_step)
        assert mock_finish.call_args.args == (process_step,)
        assert process_step.process.finished is not None
        assert process_step.process.instance.first_name == old_name
        assert process_step.process.state == Process.ProcessState.FINISHED



from .fields import (
    AdditionalResourcesField,
    BooleanField,
    CharField,
    ChoiceField,
    CodeField,
    ColorPickerField,
    DateField,
    DateRangeField,
    DateTimeField,
    DateTimeRangeField,
    DecimalField,
    DecimalRangeField,
    DynamicButtonField,
    EmojiRatingField,
    FileField,
    FloatField,
    FSMStatusField,
    HyperlinkField,
    ImageField,
    ImageURLField,
    IntegerField,
    YearField,
    JSONField,
    JSONTableField,
    JSONTextEditorField,
    ListField,
    ListSerializer,
    MarkdownTextField,
    MultipleChoiceField,
    PrimaryKeyCharField,
    PrimaryKeyField,
    PrimaryKeyRelatedField,
    RangeSelectField,
    ReadOnlyField,
    SlugRelatedField,
    ReturnContentType,
    SerializerMethodField,
    SparklineField,
    StarRatingField,
    StringRelatedField,
    TelephoneField,
    TemplatedJSONTextEditor,
    TextAreaField,
    TextField,
    TimeField,
    TimeZoneField,
    URLField,
    WBCoreType,
    decorator,
    percent_decorator,
    register_dynamic_button,
    register_only_instance_dynamic_button,
    register_only_instance_resource,
    register_resource,
)
from .serializers import ModelSerializer, RepresentationSerializer, Serializer
from .utils import (
    CurrentUserDefault,
    DefaultAttributeFromRemoteField,
    DefaultFromGET,
    DefaultFromKwargs,
    DefaultFromGetOrKwargs,
    DefaultAttributeFromObject,
    DefaultFromView,
)



from django.utils.translation import gettext as _
from wbcore.contrib.example_app.utils import get_event_types_for_league
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.instance_display.utils import (
    split_list_into_grid_template_area_sublists,
)
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

LEAGUE_FIELDS = [
    dp.Field(key="name", label=_("Name")),
    dp.Field(key="sport", label=_("Sport")),
    dp.Field(key="points_per_win", label=_("Points Per Win")),
    dp.Field(key="points_per_draw", label=_("Points Per Draw")),
    dp.Field(key="points_per_loss", label=_("Points Per Loss")),
    dp.Field(key="country", label=_("Country")),
    dp.Field(key="established_date", label=_("Established")),
    dp.Field(key="commissioner", label=_("Commissioner")),
    dp.Field(key="website", label=_("Website")),
]


class LeagueDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=LEAGUE_FIELDS)

    def get_instance_display(self) -> Display:
        matches_section = Section(
            key="matches_section",
            collapsible=False,
            title=_("Recent Matches"),
            display=Display(
                pages=[
                    Page(
                        title=_("Recent Matches"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["matches_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="matches_inline", endpoint="recent_matches")],
                            )
                        },
                    ),
                ]
            ),
        )
        table_section = Section(
            key="table_section",
            collapsible=False,
            title=_("Table"),
            display=Display(
                pages=[
                    Page(
                        title=_("Table"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["table_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="table_inline", endpoint="table")],
                            )
                        },
                    ),
                ]
            ),
        )

        if league_id := self.view.kwargs.get("pk"):
            event_types = get_event_types_for_league(int(league_id))
            player_sections = [
                Section(
                    key=f"player_{event_type['slugified_name']}_section",
                    collapsible=False,
                    title=f"{event_type['name']}s",
                    display=Display(
                        pages=[
                            Page(
                                title=f"{event_type['name']}s",
                                layouts={
                                    default(): Layout(
                                        grid_template_areas=[[f"player_{event_type['slugified_name']}_inline"]],
                                        grid_template_columns=[
                                            "minmax(min-content, 1fr)",
                                        ],
                                        grid_auto_rows=Style.MIN_CONTENT,
                                        inlines=[
                                            Inline(
                                                key=f"player_{event_type['slugified_name']}_inline",
                                                endpoint=f"player_{event_type['slugified_name']}",
                                            )
                                        ],
                                    )
                                },
                            ),
                        ]
                    ),
                )
                for event_type in event_types
            ]
            team_sections = [
                Section(
                    key=f"team_{event_type['slugified_name']}_section",
                    collapsible=False,
                    title=f"{event_type['name']}s",
                    display=Display(
                        pages=[
                            Page(
                                title=f"{event_type['name']}s",
                                layouts={
                                    default(): Layout(
                                        grid_template_areas=[[f"team_{event_type['slugified_name']}_inline"]],
                                        grid_template_columns=[
                                            "minmax(min-content, 1fr)",
                                        ],
                                        grid_auto_rows=Style.MIN_CONTENT,
                                        inlines=[
                                            Inline(
                                                key=f"team_{event_type['slugified_name']}_inline",
                                                endpoint=f"team_{event_type['slugified_name']}",
                                            )
                                        ],
                                    )
                                },
                            ),
                        ]
                    ),
                )
                for event_type in event_types
            ]
            player_section_keys = [section.key for section in player_sections]
            team_section_keys = [section.key for section in team_sections]

            return Display(
                pages=[
                    Page(
                        title=_("Main Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[
                                    ["name", "name", "sport", "matches_section", "matches_section"],
                                    [
                                        "points_per_win",
                                        "points_per_draw",
                                        "points_per_loss",
                                        "matches_section",
                                        "matches_section",
                                    ],
                                    ["country", "commissioner", ".", "matches_section", "matches_section"],
                                    ["established_date", "website", ".", "matches_section", "matches_section"],
                                    [
                                        "table_section",
                                        "table_section",
                                        "table_section",
                                        "table_section",
                                        "table_section",
                                    ],
                                ],
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(5),
                                sections=[matches_section, table_section],
                            ),
                            lte(1000): Layout(
                                grid_template_areas=[
                                    ["name", "sport"],
                                    ["points_per_win", "points_per_draw"],
                                    ["points_per_loss", "website"],
                                    ["country", "commissioner"],
                                    ["established_date", "."],
                                    ["table_section", "table_section"],
                                    ["matches_section", "matches_section"],
                                ],
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(3),
                                sections=[matches_section, table_section],
                            ),
                            lte(450): Layout(
                                grid_template_areas=[
                                    ["name"],
                                    ["sport"],
                                    ["points_per_win"],
                                    ["points_per_draw"],
                                    ["points_per_loss"],
                                    ["country"],
                                    ["commissioner"],
                                    ["website"],
                                    ["established_date"],
                                ],
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                            ),
                        },
                    ),
                    Page(
                        title=_("Player Statistics"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(
                                    player_section_keys, 3
                                ),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(5),
                                sections=player_sections,
                            ),
                            lte(1000): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(
                                    player_section_keys, 2
                                ),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(3),
                                sections=player_sections,
                            ),
                            lte(650): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(
                                    player_section_keys, 1
                                ),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                sections=player_sections,
                            ),
                        },
                    ),
                    Page(
                        title=_("Team Statistics"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(team_section_keys, 3),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(5),
                                sections=team_sections,
                            ),
                            lte(1000): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(team_section_keys, 2),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                column_gap=Style.rem(3),
                                sections=team_sections,
                            ),
                            lte(650): Layout(
                                grid_template_areas=split_list_into_grid_template_area_sublists(team_section_keys, 1),
                                grid_auto_columns="minmax(min-content, 1fr)",
                                grid_auto_rows=Style.MIN_CONTENT,
                                sections=team_sections,
                            ),
                        },
                    ),
                ]
            )
        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "name", "sport"],
                                ["points_per_win", "points_per_draw", "points_per_loss"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["name", "sport"],
                                ["points_per_win", "points_per_draw"],
                                ["points_per_loss", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["sport"],
                                ["points_per_win"],
                                ["points_per_draw"],
                                ["points_per_loss"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Details"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["country", "commissioner"],
                                ["established_date", "website"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(700): Layout(
                            grid_template_areas=[
                                ["country"],
                                ["commissioner"],
                                ["established_date"],
                                ["website"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )


class LeagueSportDisplayConfig(LeagueDisplayConfig):
    def get_list_display(self) -> ListDisplay:
        fields = LEAGUE_FIELDS.copy()
        fields.pop(1)
        return dp.ListDisplay(fields=fields)

    def get_instance_display(self) -> Display:
        if "pk" in self.view.kwargs:
            return super().get_instance_display()

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["name", "points_per_win"], ["points_per_draw", "points_per_loss"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["points_per_win"],
                                ["points_per_draw"],
                                ["points_per_loss"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Details"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["country", "commissioner"],
                                ["established_date", "website"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(700): Layout(
                            grid_template_areas=[
                                ["country"],
                                ["commissioner"],
                                ["established_date"],
                                ["website"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )



import csv
from functools import partial
from io import StringIO

import pandas as pd
from django import forms
from django.contrib import admin
from django.contrib.admin.checks import InlineModelAdminChecks
from django.contrib.admin.options import flatten_fieldsets
from django.forms import ALL_FIELDS
from django.forms.models import modelform_defines_fields
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.urls import path
from django.utils.translation import gettext_lazy as _

from .content_type.admin import ContentTypeModelAdmin  # noqa: F401
from .forms import NonrelatedInlineFormSet, nonrelated_inlineformset_factory
from .markdown.admin import AssetModelAdmin  # noqa: F401
from .models import AppliedPreset, FrontendUserConfiguration, Preset
from .release_notes.admin import ReleaseNoteModelAdmin  # noqa: F401

admin.site.register(AppliedPreset)
admin.site.register(Preset)


class FrontendUserConfigurationInline(admin.TabularInline):
    model = FrontendUserConfiguration
    extra = 0
    fields = ["id", "user", "config"]
    readonly_fields = ["id", "user", "config"]


@admin.register(FrontendUserConfiguration)
class FrontendUserConfigurationModelAdmin(admin.ModelAdmin):
    list_display = ["id", "user"]
    search_fields = ["user__email", "user__username", "id"]
    fieldsets = ((_("Main Information"), {"fields": ("user", "parent_configuration", "config")}),)
    inlines = [FrontendUserConfigurationInline]
    list_filter = ["user"]


class CsvImportForm(forms.Form):
    csv_file = forms.FileField()


class ExportCsvMixin:
    def export_as_csv(self, admin, request, queryset, *args, **kwargs):
        meta = queryset.model._meta
        field_names = [field.name for field in meta.fields]

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(meta)
        writer = csv.writer(response)

        writer.writerow(field_names)
        for obj in queryset:
            writer.writerow([getattr(obj, field) for field in field_names])

        return response

    def get_actions(self, request):
        actions = super().get_actions(request)
        actions["export_as_csv"] = (self.export_as_csv, "export_as_csv", "Export Selected")
        return actions


class ImportCsvMixin:
    change_list_template = "wbcore/admin/change_list.html"

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [path("import-csv/", self._import_csv)]
        return my_urls + urls

    def manipulate_df(self, df):
        return df

    def process_model(self, model):
        self.model.create(**model)

    def get_import_fields(self):
        return [f.name for f in self.model._meta.get_fields()]

    def _import_csv(self, request):
        if request.method == "POST":
            csv_file = request.FILES["csv_file"]

            str_text = ""
            for line in csv_file:
                str_text = str_text + line.decode()
            # Import csv as df
            df = pd.read_csv(StringIO(str_text))
            # Sanitize dataframe
            df = df.where(pd.notnull(df), None)
            df = df.drop(df.columns.difference(self.get_import_fields()), axis=1)

            # Overide this function if there is foreign key ids in the dataframe
            df = self.manipulate_df(df)
            errors = 0
            for model in df.to_dict("records"):
                # by default, process the modela as a create request. Can be override to change the behavior
                try:
                    self.process_model(model)
                except Exception:
                    errors += 1
            self.message_user(
                request,
                _("Your CSV file has been imported ( {imports}  imported, {errors} errors)").format(
                    imports=df.shape[0] - errors, errors=errors
                ),
            )
            return redirect("..")
        form = CsvImportForm()
        payload = {"form": form}
        return render(request, "wbcore/admin/csv_form.html", payload)


class NonrelatedInlineModelAdminChecks(InlineModelAdminChecks):
    """
    Check used by the admin system to determine whether or not an inline model
    has a relationship to the parent object.
    In this case we always want this check to pass.
    """

    def _check_exclude_of_parent_model(self, obj, parent_model):
        return []

    def _check_relation(self, obj, parent_model):
        return []


class NonrelatedInlineMixin:
    """
    Mixin for models not explicitly related to the inline model.
    """

    checks_class = NonrelatedInlineModelAdminChecks
    formset = NonrelatedInlineFormSet

    def get_form_queryset(self, obj):
        raise NotImplementedError()

    def save_new_instance(self, parent, instance):
        raise NotImplementedError()

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)
        for instance in instances:
            self.update_instance(formset.instance, instance)
            instance.save()
        formset.save_m2m()

    def get_formset(self, request, obj=None, **kwargs):
        if "fields" in kwargs:
            fields = kwargs.pop("fields")
        else:
            fields = flatten_fieldsets(self.get_fieldsets(request, obj))

        exclude = [*(self.exclude or []), *self.get_readonly_fields(request, obj)]
        if self.exclude is None and hasattr(self.form, "_meta") and self.form._meta.exclude:
            exclude.extend(self.form._meta.exclude)
        exclude = exclude or None

        can_delete = self.can_delete and self.has_delete_permission(request, obj)

        queryset = self.model.objects.none()
        if obj:
            queryset = self.get_form_queryset(obj)

        defaults = {
            "form": self.form,
            "formfield_callback": partial(self.formfield_for_dbfield, request=request),
            "formset": self.formset,
            "extra": self.get_extra(request, obj),
            "can_delete": can_delete,
            "can_order": False,
            "fields": fields,
            "min_num": self.get_min_num(request, obj),
            "max_num": self.get_max_num(request, obj),
            "exclude": exclude,
            "queryset": queryset,
            **kwargs,
        }

        if defaults["fields"] is None and not modelform_defines_fields(defaults["form"]):
            defaults["fields"] = ALL_FIELDS

        return nonrelated_inlineformset_factory(self.model, save_new_instance=self.save_new_instance, **defaults)


class NonrelatedStackedInline(NonrelatedInlineMixin, admin.StackedInline):
    pass


class NonrelatedTabularInline(NonrelatedInlineMixin, admin.TabularInline):
    pass



from django.conf import settings
from django.utils.module_loading import import_string
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView


class Profile(APIView):
    """Profile view to get the default user name, email and profile

    This view should be substituted by one from the authentication module
    """

    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(
            {
                "config": reverse("wbcore:frontenduserconfiguration-list", request=request),
                "name": import_string(settings.WBCORE_DEFAULT_USER_NAME)(request.user),
                "email": request.user.email,
                "profile": import_string(settings.WBCORE_PROFILE)(request),
            }
        )



import typing
from contextlib import suppress
from dataclasses import dataclass
from typing import Callable, Generator

from django.core.cache import cache
from dynamic_preferences.models import global_preferences_registry
from rest_framework.request import Request
from rest_framework.test import APIRequestFactory, force_authenticate
from wbcore.contrib.authentication.models import User
from wbcore.filters.defaults import RequiredFilterMissing

if typing.TYPE_CHECKING:
    from wbcore.pandas.views import PandasAPIViewSet
    from wbcore.viewsets import ChartViewSet


@dataclass
class CachedClass:
    view_class: type["PandasAPIViewSet"] | type["ChartViewSet"]
    view_kwargs: list[dict[str, str]] | Callable | None = None
    get_parameters: list[dict[str, str]] | Callable | None = None

    def _get_requests(self) -> Generator[Request, None, None]:
        """
        Given the stored list of GET parameters, returns a list of instantiated request with the system user as request.user

        Returns:
            A generator of DRF Request
        """
        system_user_email = global_preferences_registry.manager()["wbcore__system_user_email"]
        get_parameters = self.get_parameters if self.get_parameters else [{}]
        if callable(self.get_parameters):
            get_parameters = get_parameters()

        for get_parameter in get_parameters:
            request = APIRequestFactory().get("", data=get_parameter)
            user = User.objects.get(email=system_user_email)
            force_authenticate(request, user=user)
            request.user = user
            yield Request(request)

    def fetch_cache(self) -> typing.Any:
        """
        For every kwargs and requests, reset the cache and get the view dataframe. This hits under the hood the caching mechanism

        Returns:
            A list of generated dataframe
        """
        view_kwargs = self.view_kwargs if self.view_kwargs else [{}]
        if callable(self.view_kwargs):
            view_kwargs = view_kwargs()
        res = []
        for request in self._get_requests():
            for kwargs in view_kwargs:
                with suppress(RequiredFilterMissing):
                    view = self.view_class(request=request, kwargs=kwargs)
                    setattr(request, "parser_context", {"request": request, "view": view, "kwargs": kwargs})
                    cache_key = view._get_cache_key()
                    cache.delete(cache_key)
                    res.append(view._get_dataframe())
        return res


class PeriodicCacheRegistry:
    """
    Registry to hold the api view classes that needs to recompute their cache periodically
    """

    classes = list[CachedClass]

    def __init__(self):
        self.classes = []

    def add(
        self,
        view_class: type["PandasAPIViewSet"] | type["ChartViewSet"],
        view_kwargs: list[dict[str, str]] | Callable | None = None,
        get_parameters: list[dict[str, str]] | Callable | None = None,
    ):
        self.classes.append(
            CachedClass(
                view_class=view_class,
                view_kwargs=view_kwargs,
                get_parameters=get_parameters,
            )
        )


periodic_cache_registry = PeriodicCacheRegistry()



from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import UniqueConstraint
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from django_fsm import FSMField, transition
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin, PrimaryMixin


class ContactLocationChoices(models.TextChoices):
    WORK = "WORK", _("Work")
    PRIVATE = "PRIVATE", _("Private")
    HOME = "HOME", pgettext_lazy("Indicates the contact location", "Home")
    OTHER = "OTHER", _("Other")


class ContactTelephoneChoices(models.TextChoices):
    FIX = "FIX", _("Fix")
    CELL = "CELL", _("Cell")
    FAX = "FAX", _("Fax")
    OTHER = "OTHER", _("Other")


class BankingContact(PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    class Status(models.TextChoices):
        DRAFT = "DRAFT", _("Draft")
        PENDING = "PENDING", _("Pending")
        APPROVED = "APPROVED", _("Approved")

        @classmethod
        def get_color_map(cls):
            colors = [WBColor.RED_LIGHT.value, WBColor.YELLOW_LIGHT.value, WBColor.GREEN_LIGHT.value]
            return [choice for choice in zip(cls, colors)]

    status = FSMField(
        default=Status.DRAFT,
        choices=Status.choices,
        verbose_name=_("Status"),
    )

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.PENDING,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:bankingcontact",),
                key="submit",
                label=_("Submit"),
                action_label=_("Submit"),
                icon=WBIcon.SEND.icon,
                description_fields=_("Are you sure you want to submit this banking contact for review?"),
            )
        },
    )
    def submit(self, **kwargs):
        msg = _("<p>A banking contact is in need of approving.</p>")
        title = _("New pending banking contact")
        self.notify(title, msg)

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("directory.administrate_bankingcontact"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:bankingcontact",),
                key="approve",
                label=_("Approve"),
                action_label=_("Approve"),
                icon=WBIcon.CONFIRM.icon,
                description_fields=_("Are you sure you want to approve this banking contact?"),
            )
        },
    )
    def approve(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.DRAFT,
        permission=lambda instance, user: user.has_perm("directory.administrate_bankingcontact"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:bankingcontact",),
                key="deny",
                label=_("Deny"),
                action_label=_("Deny"),
                icon=WBIcon.REJECT.icon,
                description_fields=_("Are you sure you want to deny this banking contact?"),
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.DRAFT,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:bankingcontact",),
                key="edit",
                label=_("Edit"),
                action_label=_("Edit"),
                icon=WBIcon.EDIT.icon,
                description_fields=_("Are you sure you want to edit this banking contact?"),
            )
        },
    )
    def edit(self, by=None, description=None, **kwargs):
        pass

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for banking contact", "Primary")
    )
    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )

    entry = models.ForeignKey(
        "Entry",
        related_name="banking",
        on_delete=models.SET_NULL,
        verbose_name=pgettext_lazy("Banking Entry", "Entry"),
        null=True,
        blank=True,
    )
    institute = models.CharField(max_length=255, verbose_name=_("Institute"))
    institute_additional = models.CharField(
        max_length=255, blank=True, null=True, verbose_name=_("Institute (Additional)")
    )
    iban = models.CharField(max_length=38, blank=True, null=True, verbose_name=_("IBAN"))
    swift_bic = models.CharField(max_length=11, blank=True, null=True, verbose_name=_("SWIFT/BIC"))
    edited = models.DateTimeField(auto_now=True, verbose_name=_("Edited"))
    currency = models.ForeignKey(
        "currency.Currency",
        null=True,
        blank=True,
        db_column="linked_currency",
        on_delete=models.SET_NULL,
        verbose_name=_("Currency"),
    )

    additional_information = models.TextField(
        blank=True,
        null=True,
        verbose_name=_("Additional Information"),
        help_text=_("Can be used in place of IBAN to communicate banking information"),
    )

    # Access does nothing but signals other modules that someone has further access to this bank account
    access = models.ManyToManyField(
        to="directory.Person",
        related_name="access_to_bank_accounts",
        blank=True,
    )

    def __str__(self):
        return "%s, %s" % (self.institute, self.iban)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:bankingcontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:bankingcontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{_entry.computed_str}} {{iban}} {{institute}}"

    def notify(self, title: str, msg: str):
        """
        Get a message and a title and create a notification object for the user handling the banking contact verification.
        Args:
            title (str): The Notification title
            msg (str): The Notification message
        """

        user_qs = get_user_model().objects.filter(
            models.Q(user_permissions__codename__in=["administrate_bankingcontact"])
            | models.Q(groups__permissions__codename="administrate_bankingcontact")
        )
        for user in user_qs.all().distinct():
            send_notification(
                code="directory.banking_contact.approval",
                title=title,
                body=msg,
                user=user,
                reverse_name="wbcore:directory:bankingcontact-detail",
                reverse_args=[self.id],
            )

    class Meta:
        verbose_name = _("Banking Contact")
        verbose_name_plural = _("Banking Contacts")
        permissions = (("administrate_bankingcontact", "Can Administrate Banking Contact"),)
        constraints = [
            UniqueConstraint(name="unique_iban_entry", fields=["iban", "entry"]),
        ]

        notification_types = [
            create_notification_type(
                code="directory.banking+contact.approval",
                title="Banking Contact Notification",
                help_text="Sends out a notification when you want need to approve a change in bank details",
            )
        ]


class AddressContact(PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for address contact", "Primary")
    )
    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )
    entry = models.ForeignKey(
        "Entry", related_name="addresses", on_delete=models.SET_NULL, verbose_name=_("Entry"), null=True, blank=True
    )
    street = models.CharField(max_length=255, null=True, blank=True, verbose_name=_("Street"))
    street_additional = models.CharField(max_length=255, null=True, blank=True, verbose_name=_("Street (Additional)"))
    zip = models.CharField(max_length=255, null=True, blank=True, verbose_name=_("ZIP"))
    city = models.CharField(max_length=255, verbose_name=_("City"), null=True, blank=True)
    province = models.CharField(max_length=255, blank=True, null=True, verbose_name=_("Province"))

    geography_city = models.ForeignKey(
        "geography.Geography",
        limit_choices_to={"level": 3},
        related_name="contact_city",
        on_delete=models.PROTECT,
        verbose_name=_("City"),
        null=True,
        blank=True,
    )

    def __str__(self):
        if self.geography_city:
            return "%s, %s %s, %s" % (
                self.street,
                self.zip,
                self.geography_city.name,
                self.geography_city.parent.parent.name,
            )
        else:
            return "%s, %s" % (
                self.street,
                self.zip,
            )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:addresscontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:addresscontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{street}}, {{_geography_city.representation}}"

    class Meta:
        verbose_name = _("Address Contact")
        verbose_name_plural = _("Address Contacts")


class TelephoneContact(PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for telephone contact", "Primary")
    )
    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )
    entry = models.ForeignKey(
        "Entry", related_name="telephones", on_delete=models.SET_NULL, verbose_name=_("Entry"), null=True, blank=True
    )
    number = models.CharField(max_length=255, verbose_name=_("Number"))
    telephone_type = models.CharField(
        max_length=32,
        default=ContactTelephoneChoices.FIX,
        choices=ContactTelephoneChoices.choices,
        verbose_name=_("Type"),
    )

    def __str__(self):
        return self.number

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:telephonecontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:telephonecontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{number}}"

    class Meta:
        verbose_name = _("Telephone Contact")
        verbose_name_plural = _("Telephone Contacts")
        constraints = [
            UniqueConstraint(name="unique_number_entry", fields=["number", "entry"]),
        ]

    @classmethod
    def set_entry_primary_telephone(cls, entry, number):
        primaries = entry.telephones.filter(primary=True)
        duplicates = entry.telephones.filter(number=number)
        if primaries.exists():
            if duplicates.exists():
                entry = entry.telephones.get(id=duplicates.first().id)
                entry.primary = True
                entry.save()
            else:
                entry = entry.telephones.get(id=primaries.first().id)
                entry.number = number
                entry.save()
        else:
            if duplicates.exists():
                entry = entry.telephones.get(id=duplicates.first().id)
                entry.primary = True
                entry.save()
            else:
                cls.objects.create(entry=entry, number=number, primary=True)


class EmailContact(ComplexToStringMixin, PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for email contact", "Primary")
    )
    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )

    entry = models.ForeignKey(
        "Entry", related_name="emails", on_delete=models.SET_NULL, verbose_name=_("Entry"), null=True, blank=True
    )

    address = models.EmailField(verbose_name=_("Email Address"))

    def compute_str(self) -> str:
        repr = self.address
        if self.entry:
            repr += f" - {self.entry.computed_str}"
        return repr

    def save(self, *args, **kwargs):
        if self.address:
            self.address = self.address.lower()
        return super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:emailcontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:emailcontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    class Meta:
        verbose_name = _("E-Mail Contact")
        verbose_name_plural = _("E-Mail Contacts")
        constraints = [
            models.UniqueConstraint(fields=["entry", "address"], name="entry_address_unique_together"),
        ]

    @classmethod
    def set_entry_primary_email(cls, entry, address):
        primaries = entry.emails.filter(primary=True)
        duplicates = entry.emails.filter(address=address.lower())
        if primaries.exists():
            if duplicates.exists():
                entry = entry.emails.get(id=duplicates.first().id)
                entry.primary = True
                entry.save()
            else:
                entry = entry.emails.get(id=primaries.first().id)
                entry.address = address
                entry.save()
        else:
            if duplicates.exists():
                entry = entry.emails.get(id=duplicates.first().id)
                entry.primary = True
                entry.save()
            else:
                cls.objects.create(entry=entry, address=address, primary=True)


class WebsiteContact(PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for website contact", "Primary")
    )
    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )

    entry = models.ForeignKey(
        "Entry", related_name="websites", on_delete=models.SET_NULL, verbose_name=_("Entry"), null=True, blank=True
    )
    url = models.URLField(verbose_name=_("URL"))

    def __str__(self):
        return self.url

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:websitecontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:websitecontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{address}}"

    class Meta:
        verbose_name = _("Website Contact")
        verbose_name_plural = _("Website Contacts")
        constraints = [
            UniqueConstraint(name="unique_website_entry", fields=["url", "entry"]),
        ]


class SocialMediaContact(PrimaryMixin, WBModel):
    PRIMARY_ATTR_FIELDS = ["entry"]

    class Platforms(models.TextChoices):
        FACEBOOK = "FACEBOOK", "Facebook"
        INSTAGRAM = "INSTAGRAM", "Instagram"
        LINKEDIN = "LINKEDIN", "LinkedIn"
        REDDIT = "REDDIT", "Reddit"
        TIKTOK = "TIKTOK", "TikTok"
        TUMBLR = "TUMBLR", "Tumblr"
        TWITTER = "TWITTER", "Twitter"
        XING = "XING", "Xing"

    primary = models.BooleanField(
        default=False, verbose_name=pgettext_lazy("Primary flag for social media contact", "Primary")
    )

    location = models.CharField(
        max_length=32,
        default=ContactLocationChoices.WORK,
        choices=ContactLocationChoices.choices,
        verbose_name=_("Location"),
    )

    platform = models.CharField(
        max_length=32,
        choices=Platforms.choices,
        verbose_name=_("Platform"),
        help_text=_("The social media platform for this contact"),
    )

    entry = models.ForeignKey(
        "Entry",
        related_name="social_media",
        on_delete=models.CASCADE,
        verbose_name=_("Entry"),
        null=True,
        blank=True,
    )
    url = models.URLField(null=False, blank=False, verbose_name=_("URL"))

    def __str__(self):
        return self.url

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:socialmediacontact"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:socialmediacontactrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{platform}} - {{entry}}"

    class Meta:
        verbose_name = _("Social Media Contact")
        verbose_name_plural = _("Social Media Contacts")
        constraints = [
            UniqueConstraint(name="unique_social_entry", fields=["url", "entry"]),
        ]



import os
from datetime import timedelta
from typing import Dict

from configurations import values
from django.conf import settings
from rest_framework.request import Request
from rest_framework.reverse import reverse


class AuthenticationConfigurationMixin:
    WBCORE_DEFAULT_USER_NAME = "wbcore.contrib.authentication.configurations.get_user_name"
    WBCORE_DEFAULT_AUTH_CONFIG = "wbcore.contrib.authentication.configurations.auth_method"
    AUTH_USER_MODEL = "authentication.User"
    DEFAULT_FROM_EMAIL = values.Value("no-reply@stainly-bench.com", environ_prefix=None)


def resolve_profile(request: Request) -> Dict:
    """Returns two endpoints, one for the profile image and one for the profile widget"""

    return {
        "image": "https://image.freepik.com/free-vector/businessman-profile-cartoon_18591-58479.jpg",
        "endpoint": reverse("wbcore:authentication:userprofile-detail", args=[request.user.id], request=request),
    }


def auth_method(request):
    from wbcore.contrib.authentication.authentication import get_dev_user_from_settings

    access = settings.SIMPLE_JWT.get("ACCESS_TOKEN_LIFETIME", timedelta(minutes=5))
    refresh = settings.SIMPLE_JWT.get("REFRESH_TOKEN_LIFETIME", timedelta(days=1))

    config = {
        "type": "JWT",
        "config": {
            "token": reverse("wbcore:authentication:token_obtain_pair", request=request),
            "refresh": reverse("wbcore:authentication:token_refresh", request=request),
            "verify": reverse("wbcore:authentication:token_verify", request=request),
            "token_lifetime": {"access": access, "refresh": refresh},
            "username_field_key": "email",
            "username_field_label": "E-Mail",
            "cookie_key": settings.JWT_COOKIE_KEY,
        },
    }
    if users := get_dev_user_from_settings("email"):
        config["dev_users"] = users

    if "WB_SHOW_RESET_PASSWORD" in os.environ:
        config["config"]["reset_password"] = {
            "endpoint": reverse("wbcore:authentication:reset_password_email", request=request),
            "method": "post",
            "key": "email",
            "success_message": "An e-mail has been sent to your address with instructions to reset your password.",
        }

    return config


def get_user_name(user):
    return f"{user.profile.first_name} {user.profile.last_name}"



from enum import Enum


class WBCoreType(Enum):
    TEXT = "text"
    TELEPHONE = "tel"
    TEXTEDITOR = "texteditor"
    TEMPLATED_TEXTEDITOR = "templatedtexteditor"
    TEXTAREA = "textarea"
    MARKDOWNEDITOR = "markdowneditor"
    NUMBER = "number"
    DATETIME = "datetime"
    DATE = "date"
    DATERANGE = "daterange"
    DURATION = "duration"
    TIME = "time"
    PRIMARY_KEY = "primary_key"
    BOOLEAN = "boolean"
    SELECT = "select"
    IMAGE = "image"
    FILE = "file"
    LIST = "list"
    PERCENT = "percent"
    JSON = "json"
    JSONTABLE = "table"
    ICON = "iconselect"
    COLOR = "color"
    URL = "url"
    SPARKLINE = "sparkline"


class ReturnContentType(Enum):
    HTML = "html"
    JSON = "json"
    MARKDOWN = "markdown"


class DisplayMode(Enum):
    DECIMAL = "decimal"
    SHORTENED = "shortened"
    SCIENTIFIC = "scientific"



from typing import Iterable

import django_filters
from django.core.exceptions import ObjectDoesNotExist
from django.forms.widgets import SelectMultiple, TextInput
from django.utils.http import urlencode
from django_filters.fields import ModelChoiceField, ModelMultipleChoiceField
from rest_framework.exceptions import ParseError
from rest_framework.reverse import reverse
from wbcore.filters.mixins import WBCoreFilterMixin


class ModelChoiceFilterMixin(WBCoreFilterMixin):
    MULTIPLE: bool = False

    @classmethod
    def get_parsed_values(cls, queryset, value_ids: int | str | Iterable[int]):
        if isinstance(value_ids, str):
            value_ids = value_ids.split(",")
        elif isinstance(value_ids, int):
            value_ids = [value_ids]

        if isinstance(value_ids, list):
            for value_id in value_ids:
                try:
                    yield {
                        "value": value_id,
                        "label": str(queryset.get(id=value_id)),
                    }
                except ObjectDoesNotExist:
                    raise ParseError("Filter value invalid")

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["multiple"] = self.MULTIPLE

        queryset = self.get_queryset(request)

        if hasattr(queryset.model, "get_label_key"):
            label_key = queryset.model.get_label_key()
        else:
            label_key = self.label_key

        url = reverse(self.endpoint, request=request)
        if self.filter_params:
            if callable(self.filter_params):
                url += f"?{urlencode(self.filter_params(request, view))}"
            else:
                url += f"?{urlencode(self.filter_params)}"
        lookup_expr["input_properties"]["endpoint"] = {
            "url": url,
            "value_key": self.value_key,
            "label_key": label_key,
        }

        if default_ids := lookup_expr["input_properties"].get("default", None):
            values = self.get_parsed_values(queryset, default_ids)
            lookup_expr["input_properties"]["default"] = list(values) if self.MULTIPLE else next(values)

        if initial_ids := lookup_expr["input_properties"].get("initial", None):
            values = self.get_parsed_values(queryset, initial_ids)
            lookup_expr["input_properties"]["initial"] = list(values) if self.MULTIPLE else next(values)

        return representation, lookup_expr


class ModelMultipleChoiceFilter(ModelChoiceFilterMixin, django_filters.ModelMultipleChoiceFilter):
    class SimpleModelMultipleChoiceField(ModelMultipleChoiceField):
        """
        field class that define a simple text input as widget (instead of select). This is necessary in order to use the
        browsable api for model with a lot of items. Without it, the widget would load all the queryset option and will probably destroy the performance
        """

        class SimpleSelectMultiple(SelectMultiple):
            template_name = "django/forms/widgets/text.html"

        widget = SimpleSelectMultiple

    MULTIPLE: bool = True
    field_class = SimpleModelMultipleChoiceField
    filter_type = "select"

    def __init__(self, *args, **kwargs):
        self.endpoint = kwargs.pop("endpoint", None)
        self.filter_params = kwargs.pop("filter_params", None)
        self.value_key = kwargs.pop("value_key", None)
        self.label_key = kwargs.pop("label_key", None)

        # TODO: This is monkeypatched. Make sure that the CSVWidget is set here and only here!
        if "widget" not in kwargs:
            kwargs["widget"] = django_filters.widgets.CSVWidget
        super().__init__(*args, **kwargs)


class ModelChoiceFilter(ModelChoiceFilterMixin, django_filters.ModelChoiceFilter):
    class SimpleModelChoiceField(ModelChoiceField):
        """
        field class that define a simple text input as widget (instead of select). This is necessary in order to use the
        browsable api for model with a lot of items. Without it, the widget would load all the queryset option and will probably destroy the performance
        """

        widget = TextInput

    field_class = SimpleModelChoiceField
    filter_type = "select"
    MULTIPLE: bool = False

    def __init__(self, *args, **kwargs):
        self.endpoint = kwargs.pop("endpoint", None)
        self.value_key = kwargs.pop("value_key", None)
        self.filter_params = kwargs.pop("filter_params", None)
        self.label_key = kwargs.pop("label_key", None)
        super().__init__(*args, **kwargs)



import importlib
import re
from functools import cached_property

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver
from django.shortcuts import get_object_or_404
from django.urls import resolve
from django.urls.exceptions import NoReverseMatch
from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.decorators import action, api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializer
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models.base import merge_as_task
from wbcore.signals.instance_buttons import add_extra_button


def parse_query_parameters_list(list_str):
    try:
        return list(set(map(lambda x: int(x), filter(None, list_str.split(",")))))
    except ValueError:
        return ""


def set_identifier(identifier):
    def decorator(func):
        func.identifier = identifier
        return func

    return decorator


@api_view(["GET"])
@permission_classes([IsAuthenticated])
def api_endpoints_root(request, format=None):
    try:
        endpoints = dict()
        for wb_endpoint in settings.WB_ENDPOINTS:
            try:
                endpoints[wb_endpoint] = reverse(f"{wb_endpoint}:api-root", request=request, format=format)
            except NoReverseMatch:
                pass

        return Response(endpoints)

    except AttributeError:
        return Response({"error": "No Endpoints specified."}, status=status.HTTP_400_BAD_REQUEST)


class CloneMixin:
    """
    This Mixin can be used to inherit the clone functionality out of the box. This mixin will register the additional resources
    and the appropriate button to allow user to clone a given instance
    """

    def get_clone_instance_display(self) -> Display | None:
        """
        Return the instance display to show in the clone action button. Default to return None (i.e. No form)
        """
        return None

    @action(methods=["PATCH"], detail=True)
    def clone(self, request, pk=None):
        instance = get_object_or_404(self.queryset.model, pk=pk)
        if (clone_fct := getattr(instance, "clone", None)) and callable(clone_fct):
            kwargs = {}
            if request.data:
                serializer = self.get_serializer_class()(data=request.data)
                serializer.is_valid(raise_exception=False)
                kwargs = serializer.data
            cloned_obj = clone_fct(user=request.user, **kwargs)
            if (
                cloned_obj
                and (clone_id := getattr(cloned_obj, "id", None))
                and (endpoint_config_class := getattr(self, "endpoint_config_class", None))
            ):
                endpoint_config = endpoint_config_class(self, request, cloned_obj)
                if endpoint := endpoint_config.get_instance_endpoint():
                    return Response(
                        {
                            "endpoint": re.sub(
                                pattern=r"[^\/]+(?=\/$|$)",
                                repl=str(clone_id),
                                string=endpoint,
                            )
                        }
                    )
        return Response({})


@receiver(add_extra_button)
def add_clone_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if instance and pk and issubclass(view.__class__, CloneMixin):
        object = view.get_object()
        if getattr(object, "is_cloneable", True) and (endpoint_basename := object.get_endpoint_basename()):
            endpoint = reverse(f"{endpoint_basename}-clone", args=[pk], request=request)
            identifier = getattr(view, "IDENTIFIER", "{0.app_label}:{0.model}".format(view.get_content_type()))
            return bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=(identifier,),
                endpoint=endpoint,
                action_label=_("Clone"),
                title=_("Clone"),
                label=_("Clone"),
                icon=WBIcon.COPY.icon,
                instance_display=view.get_clone_instance_display(),
            )


class MergeMixin:
    """
    This Mixin can be used to inherit the clone functionality out of the box. This mixin will register the additional resources
    and the appropriate button to allow user to merge a given instance
    """

    @cached_property
    def model(self):
        return self.queryset.model

    @cached_property
    def has_user_merge_permission(self) -> bool:
        return getattr(self, "MERGE_PERMISSION", f"administrate_{self.model._meta.model_name}")

    @cached_property
    def can_merge(self):
        return hasattr(self.queryset.model, "merge") and self.has_user_merge_permission

    def get_merged_object_representation_serializer(self):
        """
        Return the representation serializer used for the action button merged object field. Default to the model representation endpoint serializer
        """
        view_func = resolve(reverse(self.model.get_representation_endpoint())).func
        module = importlib.import_module(view_func.__module__)
        view = getattr(module, view_func.__name__)
        return view.serializer_class

    def get_merged_object_representation_serializer_filter_params(self) -> dict[str, str]:
        """
        Return the representation serializer filter kwargs. Default to an empty dict
        """
        return dict()

    @action(methods=["PATCH"], detail=True)
    def merge(self, request, pk=None):
        instance = get_object_or_404(self.queryset.model, pk=pk)
        merged_instance = get_object_or_404(self.queryset.model, pk=request.POST.get("merged_object", None))
        try:
            content_type = ContentType.objects.get_for_model(self.model)
        except ContentType.DoesNotExist:
            content_type = None
        if content_type and self.has_user_merge_permission and hasattr(instance, "merge"):
            merge_as_task.delay(content_type.id, instance.id, merged_instance.id)
            return Response({"status": "Merged with success"})
        return Response({}, status=400)


@receiver(add_extra_button)
def add_merge_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if instance and pk and issubclass(view.__class__, CloneMixin) and getattr(view, "can_merge", False):
        object = view.get_object()
        if getattr(object, "is_mergeable", True) and (endpoint_basename := object.get_endpoint_basename()):
            endpoint = reverse(f"{endpoint_basename}-merge", args=[pk], request=request)
            identifier = getattr(view, "IDENTIFIER", "{0.app_label}:{0.model}".format(view.get_content_type()))

            class MergeSerializer(wb_serializer.ModelSerializer):
                merged_object = wb_serializer.PrimaryKeyRelatedField(queryset=view.model.objects.all())
                _merged_object = view.get_merged_object_representation_serializer()(
                    source="merged_object",
                    filter_params=view.get_merged_object_representation_serializer_filter_params(),
                )

                class Meta:
                    model = view.model
                    fields = [
                        "id",
                        "merged_object",
                        "_merged_object",
                    ]

            return bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=(identifier,),
                endpoint=endpoint,
                action_label=_("Merge"),
                title=_("Merge"),
                label=_("Merge"),
                icon=WBIcon.MERGE.icon,
                serializer=MergeSerializer,
                instance_display=create_simple_display([["merged_object"]]),
            )



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class DataEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> str:
        endpoint: str = super().get_create_endpoint(**kwargs)
        if workflow_id := self.view.kwargs.get("workflow_id"):
            endpoint += f"?workflow={workflow_id}"
        return endpoint



from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline
from django_admin_inline_paginator.admin import TabularInlinePaginated

from .models import (
    DataBackend,
    ExportSource,
    ImportCredential,
    ImportedObjectProviderRelationship,
    ImportSource,
    ParserHandler,
    Provider,
    Source,
    import_data_as_task,
)


def reprocess_import_source(modeladmin, request, queryset):
    for import_source in queryset:
        import_data_as_task.delay(import_source.id, force_reimport=True)


def reprocess_source(modeladmin, request, queryset):
    for source in queryset:
        source.trigger_workflow()


class ImportedObjectProviderRelationshipInline(TabularInlinePaginated):
    ordering = ["object_id"]
    model = ImportedObjectProviderRelationship
    extra = 0
    readonly_fields = ("content_type", "object_id")
    raw_id_fields = ["content_type", "provider"]
    per_page = 25

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("content_type", "provider")


class ProviderGenericInline(GenericTabularInline):
    model = ImportedObjectProviderRelationship


@admin.register(ParserHandler)
class ParserHandlerModelAdmin(admin.ModelAdmin):
    list_display = ["parser", "handler"]

    search_fields = ["parser", "handler"]


@admin.register(Provider)
class ProviderModelAdmin(admin.ModelAdmin):
    list_display = ["title"]
    search_fields = ["title"]
    inlines = [ImportedObjectProviderRelationshipInline]


@admin.register(DataBackend)
class DataBackendModelAdmin(admin.ModelAdmin):
    list_display = [
        "title",
        "save_data_in_import_source",
        "backend_class_path",
        "backend_class_name",
        "provider",
    ]
    search_fields = ["title"]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("provider")


@admin.register(Source)
class SourceModelAdmin(admin.ModelAdmin):
    list_display = ["uuid", "title", "is_active", "crontab", "data_backend"]
    list_filter = ["data_backend", "is_active"]
    raw_id_fields = ["crontab", "periodic_task"]
    search_fields = ["uuid", "data_backend__title", "data_backend__provider__title", "parser_handler__parser"]
    readonly_fields = ["periodic_task"]

    actions = [reprocess_source]
    autocomplete_fields = ["parser_handler", "data_backend"]

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("data_backend", "periodic_task", "crontab")
            .prefetch_related("parser_handler", "credentials")
        )


@admin.register(ExportSource)
class ExportSourceModelAdmin(admin.ModelAdmin):
    list_display = ["status", "file", "resource_path", "content_type", "created", "last_updated", "creator"]
    list_filter = ["status", "resource_path", "content_type"]
    autocomplete_fields = ["creator", "content_type"]
    readonly_fields = ("created", "last_updated", "query_str", "query_params")
    fieldsets = (
        (
            "",
            {
                "fields": (
                    ("status", "created", "last_updated"),
                    ("content_type", "format"),
                    ("resource_path", "resource_kwargs"),
                    ("query_str", "query_params"),
                    ("file", "origin", "creator"),
                    ("data",),
                    ("log",),
                ),
            },
        ),
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("content_type")


@admin.register(ImportSource)
class ImportSourceModelAdmin(admin.ModelAdmin):
    list_display = ["status", "file", "parser_handler", "source", "created", "last_updated"]
    list_filter = ["status", "source", "parser_handler__handler", "parser_handler__parser"]
    autocomplete_fields = ["source", "parser_handler"]
    actions = [reprocess_import_source]
    readonly_fields = ("created", "last_updated")
    fieldsets = (
        (
            "",
            {
                "fields": (
                    ("status", "created", "last_updated"),
                    ("parser_handler", "source", "origin"),
                    ("file", "save_data", "progress_index"),
                    ("creator",),
                    ("data",),
                    ("log",),
                    ("errors_log",),
                ),
            },
        ),
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("source", "parser_handler")


@admin.register(ImportCredential)
class ImportCredentialModelAdmin(admin.ModelAdmin):
    list_display = ["key", "type", "username", "validity_start", "validity_end"]



# Generated by Django 5.0.6 on 2024-07-09 10:42

from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("example_app", "0003_stadium_seating_capacity_stadium_standing_capacity"),
        ("example_app", "0004_remove_stadium_capacity"),
    ]

    dependencies = [
        ("example_app", "0002_sportperson_profile"),
    ]

    operations = [
        migrations.AddField(
            model_name="stadium",
            name="seating_capacity",
            field=models.PositiveIntegerField(default=0, verbose_name="Seating Capacity"),
        ),
        migrations.AddField(
            model_name="stadium",
            name="standing_capacity",
            field=models.PositiveIntegerField(default=0, verbose_name="Standing Capacity"),
        ),
        migrations.RemoveField(
            model_name="stadium",
            name="capacity",
        ),
    ]



import re
from contextlib import suppress

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist, FieldError
from django.db.models import CharField, Value
from django.db.models.functions import Concat
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from wbcore import viewsets

from .filters import ContentTypeFilterSet
from .serializers import (
    ContentTypeRepresentationSerializer,
    DynamicObjectIDRepresentationSerializer,
    GenericModel,
)


class ContentTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    filterset_class = ContentTypeFilterSet
    queryset = ContentType.objects.all()
    serializer_class = ContentTypeRepresentationSerializer
    search_fields = ["app_label", "model"]


class DynamicObjectIDRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcore:object_id"
    queryset = GenericModel.objects.none()
    search_fields = ["label"]

    @cached_property
    def model(self):
        content_type = get_object_or_404(ContentType, pk=self.kwargs.get("pk", self.request.GET.get("content_type")))
        return content_type.model_class()

    @cached_property
    def label_keys(self) -> list[str]:
        label_keys = ["computed_str"]
        if (r := getattr(self.model, "get_representation_label_key", None)) and callable(r):
            label_keys = re.findall(r"\{\{(.*?)\}\}", r())
        validated_keys = []
        for key in label_keys:
            with suppress(FieldDoesNotExist):
                if self.model._meta.get_field(key):
                    validated_keys.append(key)
        return validated_keys

    def get_serializer_class(self):
        class ComputedStrSerializer(DynamicObjectIDRepresentationSerializer):
            class Meta:
                fields = ("id", "label", *self.label_keys)
                model = self.model

        return ComputedStrSerializer

    def filter_queryset(self, queryset):
        """
        We allow basic filtering on the model fields
        """
        filter_params = dict(self.request.GET.dict())  # convert immutable QueryDict into a mutable dict
        filter_params.pop("content_type", None)
        for key, value in filter_params.items():
            with suppress(FieldError, ValueError):
                queryset = queryset.filter(**{key: value})
        return super().filter_queryset(queryset)

    def get_queryset(self):
        try:
            if not self.label_keys:
                raise FieldError()
            concat_fields = []
            for field in self.label_keys:
                concat_fields.append(field)
                concat_fields.append(Value(" "))
            return self.model.objects.annotate(
                label=Concat(*concat_fields, output_field=CharField()),
            )
        except FieldError:
            return self.model.objects.none()



from contextlib import suppress
from typing import Any, List, Optional

from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from django.core import checks
from django.db import models
from django.db.models.base import ModelBase
from django.utils.functional import cached_property
from django_better_admin_arrayfield.models.fields import ArrayField
from igraph import Graph
from wbcore.signals.models import get_dependant_dynamic_fields_instances

from .fields import DynamicDecimalField, DynamicFloatField


class DynamicMetaClass(ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        # Construct dependencies graph
        dynamic_field_names = {
            field.name: field.dependencies
            for field in new_class._meta.fields
            if type(field) in [DynamicFloatField, DynamicDecimalField]
        }
        ordered_dependencies_fields = []
        if dynamic_field_names:
            g = Graph(len(dynamic_field_names), directed=True)
            # label vertices
            g.vs["name"] = list(dynamic_field_names.keys())
            for field_name, dependencies in dynamic_field_names.items():
                for dependency in dependencies:
                    g[dependency, field_name] = 1
            # Store dynamic field ordered by topological sorting
            ordered_dependencies_fields = [g.vs[v_id]["name"] for v_id in g.topological_sorting()]
        new_class.add_to_class("ordered_dependencies_fields", ordered_dependencies_fields)
        return new_class


class DynamicModel(models.Model, metaclass=DynamicMetaClass):
    DATE_RANGE_UPPER_FIELD_NAME = "date"
    DATE_RANGE_LOWER_FIELD_NAME = "date"

    lock_all_dynamic_fields = models.BooleanField(
        default=False, help_text="If True, will not allow dynamic fields to be updated"
    )
    locked_dynamic_fields = ArrayField(
        models.CharField(
            max_length=128,
        ),
        blank=True,
        default=list,
    )
    updated_at = models.DateTimeField(auto_now=True)

    @classmethod
    def get_static_locked_fields(cls) -> list[str]:
        """
        Override if the model needs to define a set of dynamic fields that can only be set when the value is null

        Returns:
            Return a list of field names
        """
        return []

    @cached_property
    def _locked_fields(self) -> set[str]:
        return set(self.locked_dynamic_fields + self.get_static_locked_fields())

    def set_dynamic_field(
        self,
        update_all_dynamic_fields: bool,
        update_only_fields: list[str] | None = None,
        update_exclude_fields: list[str] | None = None,
    ):
        if not update_only_fields:
            update_only_fields = []
        if not update_exclude_fields:
            update_exclude_fields = []
        # Loop over the dynamic field ordered by their place in the dependency graph
        for dynamic_field_name in self.ordered_dependencies_fields:
            # Check if field value is not assigned or if we explicitly want to override its value
            if (
                not self.lock_all_dynamic_fields
                and (
                    getattr(self, dynamic_field_name, None) is None
                    or (
                        dynamic_field_name not in self._locked_fields
                        and (update_all_dynamic_fields or dynamic_field_name in update_only_fields)
                    )
                )
            ) and dynamic_field_name not in update_exclude_fields:
                # call the appropriate callback and check if this is a function
                if (
                    (compute_function := getattr(self, f"_compute_{dynamic_field_name}", None))
                    and (callable(compute_function))
                    and (res := compute_function()) is not None
                ):
                    setattr(self, dynamic_field_name, res)

    @classmethod
    def update_dynamic_fields_of_dependant_objects(
        cls,
        updated_instance,
        already_visited_instances: set[Any] | None = None,
        update_all_dynamic_fields: bool | None = True,
    ):
        if not already_visited_instances:
            already_visited_instances = set()
        for _, dependant_instance_generator in get_dependant_dynamic_fields_instances.send(
            sender=updated_instance.__class__, instance=updated_instance
        ):
            for dependant_instance in dependant_instance_generator:
                if dependant_instance not in already_visited_instances:
                    dependant_instance.set_dynamic_field(update_all_dynamic_fields)
                    already_visited_instances.add(updated_instance)
                    DynamicModel.update_dynamic_fields_of_dependant_objects(
                        dependant_instance, already_visited_instances=already_visited_instances
                    )

    def save(
        self,
        *args,
        update_all_dynamic_fields: Optional[bool] = False,
        update_only_fields: Optional[List] = None,
        update_exclude_fields: Optional[List] = None,
        set_dynamic_field_on_save: Optional[bool] = True,
        **kwargs,
    ):
        if set_dynamic_field_on_save:
            self.set_dynamic_field(update_all_dynamic_fields, update_only_fields, update_exclude_fields)
        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


@shared_task
def set_dynamic_field_as_task(content_type_id, object_id, **kwargs):
    content_type = ContentType.objects.get(id=content_type_id)
    instance = content_type.model_class().objects.get(id=object_id)
    instance.set_dynamic_field(True, **kwargs)
    instance.save()


class WBModel(models.Model):
    """Implements checks on the model class"""

    id: int

    class Meta:
        abstract = True

    @classmethod
    def _check_model_methods(cls) -> list[checks.Error]:
        methods_to_check = [
            "get_endpoint_basename",
            "get_representation_value_key",
            "get_representation_endpoint",
            "get_representation_label_key",
        ]
        method_errors = []
        for method in methods_to_check:
            if not hasattr(cls._meta.model, method):
                method_errors.append(checks.Error(f"{cls._meta.model.__name__} does not have method {method}"))
        return method_errors

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_model_methods())
        return errors


@shared_task
def merge_as_task(content_type_id: int, main_object_id: int, merged_object_id: int):
    content_type = ContentType.objects.get(id=content_type_id)
    with suppress(content_type.model_class().DoesNotExist):
        main_object = content_type.get_object_for_this_type(id=main_object_id)
        merged_object = content_type.get_object_for_this_type(id=merged_object_id)
        if hasattr(main_object, "merge") and callable(main_object.merge):
            main_object.merge(merged_object)



import re
from contextlib import suppress

from django.core.management.base import BaseCommand
from markdown import Markdown
from wbcore.release_notes.models import ReleaseNote


def parse_changelog(change_log_path: str, changelog_section_regex: str):
    with open(change_log_path) as file:
        section = ""
        previous_matched_version = None
        previous_matched_date = None

        for line in file:
            match = re.search(changelog_section_regex, line)
            if match:
                if previous_matched_version and previous_matched_date:
                    yield previous_matched_version, previous_matched_date, section
                section = ""
                previous_matched_version = match.group(1)
                previous_matched_date = match.group(2)
            section += line + "\n"
        if previous_matched_version and previous_matched_date:
            yield previous_matched_version, previous_matched_date, section


class Command(BaseCommand):
    help = "Handles the creation/deletion/changing of release notes through markdown files"

    def add_arguments(self, parser):
        parser.add_argument("module_changelog_path_tuples", nargs="*", type=str, default=["stainly:CHANGELOG.md"])
        parser.add_argument(
            "changelog_section_regex", nargs="?", type=str, default=r"^##\sv(\d+\.\d+\.\d+).*(\d{4}-\d{2}-\d{2}).*$"
        )

    def handle(self, *args, **options):
        for module_changelog_path_tuple in options["module_changelog_path_tuples"]:
            module, changelog_path = module_changelog_path_tuple.split(":")
            stale_release_notes = ReleaseNote.objects.filter(module=module)
            with suppress(FileNotFoundError):
                for version, release_date, section in parse_changelog(
                    changelog_path, options["changelog_section_regex"]
                ):
                    md = Markdown(extensions=["meta"], tab_length=2)
                    content = md.convert(section)
                    release_note_obj, created = ReleaseNote.objects.update_or_create(
                        version=version,
                        module=module,
                        defaults={"release_date": release_date, "summary": f"v{version}", "notes": content},
                    )
                    stale_release_notes = stale_release_notes.exclude(id=release_note_obj.id)
                stale_release_notes.delete()



import pytest
from selenium.webdriver.common.action_chains import ActionChains
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.authentication.models import User
from wbcore.contrib.example_app.models import Team
from wbcore.contrib.example_app.serializers import TeamModelSerializer
from wbcore.contrib.example_app.tests.e2e import create_new_team_instance
from wbcore.test import (
    click_new_button,
    delete_list_entry,
    edit_list_instance,
    is_counter_as_expected,
    is_string_not_visible,
    is_text_visible,
    open_menu_item,
    set_up,
)

USER_PASSWORD = "User_Password"


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestTeam:
    def test_create_edit_delete_team(self, live_server, selenium):
        # Creating a test user and login to the WB
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        # -----> CREATE A NEW TEAM <----- #
        team_a = create_new_team_instance(selenium, ["name", "founded_date"], False)
        open_menu_item(selenium, "Teams", perform_mouse_move=True)

        assert is_counter_as_expected(selenium, Team.objects.count())
        assert is_text_visible(selenium, team_a.name)
        click_new_button(selenium)
        assert is_text_visible(selenium, "Create Team")
        team_b = create_new_team_instance(selenium, ["name", "founded_date"], True)

        assert is_counter_as_expected(selenium, Team.objects.count())
        assert is_text_visible(selenium, team_b.name)

        # ------> EDIT A NEW LEAGUE <------ #

        edit_list_instance(selenium, actions, team_a.name, TeamModelSerializer(team_a), {"name": "Fun Team"})
        assert is_string_not_visible(selenium, team_a.name)
        assert is_text_visible(selenium, "Fun Team")
        assert is_counter_as_expected(selenium, Team.objects.count())

        # -----> Delete <----- #

        delete_list_entry(selenium, actions, team_b.name)
        assert is_counter_as_expected(selenium, Team.objects.count())
        assert is_text_visible(selenium, "Fun Team")
        assert is_string_not_visible(selenium, team_b.name)



from datetime import date
from enum import Enum

import phonenumbers
from django.core.validators import validate_email
from django.forms import ValidationError
from django.utils import timezone
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.reverse import reverse
from slugify import slugify
from wbcore import serializers
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbcore.contrib.example_app.models import Player, Role, SportPerson, Stadium, Team
from wbcore.contrib.example_app.serializers import (
    RoleRepresentationSerializer,
    StadiumRepresentationSerializer,
)
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer
from wbcore.serializers.fields.number import DisplayMode


class TeamErrorMessages(Enum):
    team_exists = _("A team with this name already exists.")
    wrong_founding_date = _("The founding date must be in the past.")


class SportPersonModelSerializer(serializers.ModelSerializer):
    profile = serializers.PrimaryKeyRelatedField(read_only=True)
    _profile = PersonRepresentationSerializer(many=False, source="profile")
    _roles = RoleRepresentationSerializer(source="roles", many=True)

    def create(self, validated_data):
        if not validated_data.get("profile", None):
            person = Person.objects.create(
                first_name=validated_data["first_name"], last_name=validated_data["last_name"]
            )
            validated_data["profile"] = person
        return super().create(validated_data)

    class Meta:
        model = SportPerson
        fields = (
            "id",
            "first_name",
            "last_name",
            "computed_str",
            "roles",
            "_roles",
            "profile",
            "_profile",
            "profile_image",
            "_additional_resources",
        )


class SportPersonTooltipSerializer(serializers.ModelSerializer):
    class Meta:
        model = SportPerson
        fields = ("id", "first_name", "last_name", "profile_image")


class SportPersonRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:person-detail")

    class Meta:
        model = SportPerson
        fields = ("id", "computed_str", "_detail")


class TeamRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:team-detail")

    class Meta:
        model = Team
        fields = ("id", "name", "_detail")


class TeamModelSerializer(serializers.ModelSerializer):
    name = serializers.CharField(placeholder=_("Enter team name here"))
    _coach = SportPersonRepresentationSerializer(source="coach")
    home_stadium = serializers.PrimaryKeyRelatedField(
        queryset=Stadium.objects.all(), label=_("Home Stadium"), default=serializers.DefaultFromGET("stadium")
    )
    _home_stadium = StadiumRepresentationSerializer(source="home_stadium")
    _city = GeographyRepresentationSerializer(filter_params={"level": 3}, source="city")
    _opponents = TeamRepresentationSerializer(source="opponents", many=True)
    _group_key = serializers.IntegerField(read_only=True)
    phone_number = serializers.TelephoneField(label=pgettext_lazy("Phonenumber", "Number"), required=False)

    def validate(self, data: dict) -> dict:
        name: str | None = data.get("name", None)
        founded_date: date | None = data.get("founded_date", None)
        phone_number = data.get("phone_number", None)
        email = data.get("email", None)

        if name:
            team = Team.objects.filter(slugify_name=slugify(name, separator=" "))
            if obj := self.instance:
                team = team.exclude(id=obj.id)
            if team.exists():
                raise ValidationError({"name": TeamErrorMessages.team_exists.value})
        if founded_date and founded_date > timezone.now().date():
            raise ValidationError({"founded_date": TeamErrorMessages.wrong_founding_date.value})
        if email:
            try:
                validate_email(email)
            except ValidationError:
                raise ValidationError({"email": _("Invalid e-mail address")})
        if phone_number:
            try:
                if phone_number.startswith("00"):
                    phone_number = phone_number.replace("00", "+", 1)
                parser_number = phonenumbers.parse(
                    phone_number, global_preferences_registry.manager()["directory__main_country_code"]
                )
                if parser_number:
                    formatted_number = phonenumbers.format_number(parser_number, phonenumbers.PhoneNumberFormat.E164)
                    data["phone_number"] = formatted_number
            except Exception:
                raise ValidationError({"phone_number": _("Invalid phone number format")})

        return super().validate(data)

    @serializers.register_resource()
    def website(self, instance: Team, request, user):
        return {"website": instance.website}

    @serializers.register_resource()
    def coach(self, instance: Team, request, user):
        if instance.coach:
            return {
                "coach": reverse(
                    "example_app:person-detail",
                    args=[instance.coach.id],
                    request=self.context.get("request"),
                )
            }
        return {}

    @serializers.register_resource()
    def coach_tooltip(self, instance: Team, request, user):
        if instance.coach:
            return {
                "coach_tooltip": reverse(
                    "example_app:persontooltip-detail",
                    args=[instance.coach.id],
                    request=self.context.get("request"),
                )
            }
        return {}

    @serializers.register_only_instance_resource()
    def players_inline(self, instance: Team, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "players": reverse("example_app:player-team-list", args=[instance.pk], request=self.context["request"])
        }

    @serializers.register_only_instance_resource()
    def matches_inline(self, instance: Team, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "recent_matches": reverse(
                "example_app:match-team-list", args=[instance.pk], request=self.context["request"]
            )
        }

    class Meta:
        model = Team
        fields = (
            "id",
            "name",
            "slugify_name",
            "computed_str",
            "city",
            "_city",
            "coach",
            "_coach",
            "home_stadium",
            "_home_stadium",
            "opponents",
            "_opponents",
            "founded_date",
            "website",
            "phone_number",
            "email",
            "duration_since_last_win",
            "order",
            "_group_key",
            "_additional_resources",
        )
        read_only_fields = ("duration_since_last_win",)
        optional_fields = ("current_players", "former_players", "home_matches", "away_matches")


def get_player_role() -> list[Role]:
    player_role, created = Role.objects.get_or_create(title="Player")
    return [player_role]


class PlayerModelSerializer(SportPersonModelSerializer):
    roles = serializers.PrimaryKeyRelatedField(
        queryset=Role.objects.all(), label=_("Roles"), default=get_player_role, many=True
    )
    _roles = RoleRepresentationSerializer(source="roles", many=True)
    current_team = serializers.PrimaryKeyRelatedField(
        queryset=Team.objects.all(),
        label=_("Current Team"),
        default=serializers.DefaultFromGET("team"),
    )
    transfer_value = serializers.DecimalField(
        display_mode=DisplayMode.SHORTENED, precision=1, max_digits=10, decimal_places=2
    )
    game_activity = serializers.RangeSelectField(
        color="rgb(220,20,60)", label=_("Activity Heat"), read_only=True, required=False
    )
    _current_team = TeamRepresentationSerializer(source="current_team")
    _former_teams = TeamRepresentationSerializer(source="former_teams", many=True)
    _overwrites = serializers.SerializerMethodField("overwrites")
    player_strength = serializers.StarRatingField()

    def overwrites(self, obj):
        if obj:
            transfer_value = obj.transfer_value if obj.transfer_value else 0
            operator = "!=" if transfer_value < 99999 else "="
            overwrite = {
                "fields": [
                    {
                        "rule": {"key": "transfer_value", "operator": operator, "value": "number"},
                        "overwrite": {"displayMode": "decimal", "precision": 0},
                    }
                ]
            }
            return overwrite

    def validate_player_strength(self, value):
        if value < 0 or value > 5:
            raise ValidationError(_("Rating must be between 0 and 5."))
        return value

    class Meta:
        model = Player
        fields = (
            "id",
            "first_name",
            "last_name",
            "computed_str",
            "roles",
            "_roles",
            "current_team",
            "_current_team",
            "is_active",
            "is_injured",
            "former_teams",
            "_former_teams",
            "position",
            "player_strength",
            "game_activity",
            "transfer_value",
            "_overwrites",
        )


class TreeViewPlayerModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Player
        fields = ("id", "order", "computed_str", "current_team")


class PlayerRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:player-detail")

    class Meta:
        model = Player
        fields = ("id", "computed_str", "_detail")



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class CalendarItemTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Calendar")



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ShareableLinkModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="valid_until", label=_("Valid Until")),
                dp.Field(key="one_time_link", label=_("One Time Link")),
                dp.Field(key="link", label=_("Link")),
            ],
            legends=[
                dp.Legend(
                    key="valid",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Valid"),
                            value=True,
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="valid",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["link", "link"], ["valid_until", "one_time_link"], [repeat_field(2, "hits_section")]],
            [create_simple_section("hits_section", _("Hits"), [["hits"]], "hits", collapsed=False)],
        )


class ShareableLinkAccessModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="accessed", label=_("Date")),
                dp.Field(key="metadata_repr", label=_("Meta Data")),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["accessed", "metadata"]])



from django.conf import settings
from django.utils.module_loading import import_string
from rest_framework.request import Request
from wbcore.configs.decorators import register_config


@register_config
def authentication_config(request: Request) -> tuple[str, dict]:
    return "authentication", import_string(settings.WBCORE_DEFAULT_AUTH_CONFIG)(request)



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.documents.models import Document, ShareableLink, ShareableLinkAccess
from wbcore.contrib.documents.serializers import DocumentRepresentationSerializer
from wbcore.utils.strings import snake_case_to_human


class ShareableLinkModelSerializer(serializers.ModelSerializer):
    document = serializers.PrimaryKeyRelatedField(
        queryset=Document.objects.all(), label="Document", default=serializers.DefaultFromGET("document")
    )
    _document = DocumentRepresentationSerializer(source="document")
    valid = serializers.BooleanField(label=_("Valid"), default=True, read_only=True)

    @serializers.register_resource()
    def hits(self, instance, request, user):
        base_url = reverse("wbcore:documents:linkaccess-list", request=request)
        return {"hits": f"{base_url}?shareable_link={instance.id}"}

    @serializers.register_resource()
    def manually_invalidate(self, instance, request, user):
        if "view" in request.parser_context and hasattr(instance, "is_valid") and instance.is_valid():
            return {
                "manually_invalidate": reverse(
                    "wbcore:documents:link-invalidate",
                    args=[instance.id],
                    request=request,
                )
            }
        return {}

    class Meta:
        model = ShareableLink
        fields = (
            "id",
            "link",
            "document",
            "_document",
            "valid_until",
            "one_time_link",
            "manual_invalid",
            "uuid",
            "_additional_resources",
            "valid",
        )


class ShareableLinkRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:documents:linkrepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:documents:link-detail")
    _detail_preview = serializers.HyperlinkField(reverse_name="wbcore:documents:link-detail")

    class Meta:
        model = ShareableLink
        fields = ("id", "uuid", "_detail", "_detail_preview")


class ShareableLinkAccessModelSerializer(serializers.ModelSerializer):
    _shareable_link = ShareableLinkRepresentationSerializer(source="shareable_link")

    metadata_repr = serializers.SerializerMethodField(read_only=True)

    def get_metadata_repr(self, obj):
        return f"""
        <ul>
        {"".join([f"<li>{snake_case_to_human(key)}: {value}</li>" for key, value in obj.metadata.items()])}
        </ul>
        """

    class Meta:
        model = ShareableLinkAccess
        fields = (
            "id",
            "shareable_link",
            "_shareable_link",
            "metadata",
            "metadata_repr",
            "accessed",
        )


class ShareableLinkAccessRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:documents:linkaccessrepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:documents:linkaccess-detail")
    _detail_preview = serializers.HyperlinkField(reverse_name="wbcore:documents:linkaccess-detail")

    class Meta:
        model = ShareableLinkAccess
        fields = ("id", "metadata", "_detail", "_detail_preview")



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RevisionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcore:revision-list", request=self.request)

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class VersionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcore:version-list", request=self.request)

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class VersionRevisionEndpointConfig(VersionEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcore:revision-version-list", args=[self.view.kwargs["revision_id"]], request=self.request)



import re
from contextlib import suppress

import tablib
from django.template import Context, Template
from import_export import resources
from wbcore.serializers import ListSerializer
from wbcore.serializers.fields.related import ListSerializer as RelatedListSerializer
from wbcore.utils.importlib import import_from_dotted_path


class ExportResourceMixin:
    DUMMY_FIELD_MAP = {}

    def get_export_dummy_data(self) -> list[str]:
        """
        Returns a list in the same order as the headers with the fixture data that correspond to the header
        """
        res = []
        for header in self.get_export_headers():
            if value := self.DUMMY_FIELD_MAP.get(header):
                if callable(value):
                    value = value()
                res.append(value)
            else:
                res.append("")
        return res

    def export(self, queryset=None, *args, **kwargs):
        """
        override the functionality from import-export in order to allow empty queryset, which correspond to template export
        """
        if queryset is None:
            headers = self.get_export_headers()
            data = self.get_export_dummy_data()
            return tablib.Dataset(data, headers=headers)
        return super().export(*args, queryset=queryset, **kwargs)


class FilterModelResource(ExportResourceMixin, resources.ModelResource):
    """
    This class inherit from modelResource and provides few functionality out of the box:
    - Allow to define a template export and a map of field to corresponding dummy data to fill in the template with
    - Can instantiate the resource with a list of keyword argument that will be injected as field value before instance saving
    - Filter the queryset with the specified keyword argument
    """

    def __init__(self, filter_kwargs=None, request=None, **kwargs):
        super().__init__(**kwargs)
        if not filter_kwargs:
            filter_kwargs = dict()
        self.filter_kwargs = filter_kwargs
        self.request = request

    def before_save_instance(self, instance, *args):
        """
        For all specified `filter_kwargs`, set the attribute as field
        """
        for k, v in self.filter_kwargs.items():
            if not getattr(self, k, None):
                setattr(instance, k, v)

    def get_queryset(self):
        """
        Filter the queryset with the given keyword arguments
        """
        return self._meta.model.objects.filter(**self.filter_kwargs)


class ViewResource(ExportResourceMixin, resources.Resource):
    DEFAULT_PAGINATION_LIMIT: int = 200

    def __init__(
        self,
        columns_map: dict[str, str] | None = None,
        serializer_class=None,
        request=None,
        serializer_class_path: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # in that case, serializer_class is lazy loaded,
        if serializer_class_path:
            serializer_class = import_from_dotted_path(serializer_class_path)
        if not serializer_class:
            raise ValueError("ViewResource needs a serializer_class")
        self.serializer_class = serializer_class
        if not columns_map:
            columns_map = dict()
        self.columns_map = columns_map
        self.request = request

    def iter_queryset(self, queryset):
        serializer = self.serializer_class(queryset, many=True)
        return super().iter_queryset(serializer.data)

    def get_export_fields(self, selected_fields=None):
        res = []
        # Get default fields from the list display
        if "id" not in self.columns_map:
            res.append(
                resources.Field(
                    attribute="id",
                    column_name="id",
                    readonly=True,
                )
            )
        if self.columns_map:
            for key, label in self.columns_map.items():
                res.append(
                    resources.Field(
                        attribute=key,
                        column_name=label,
                        readonly=True,
                    )
                )
        else:
            for field_name, field in self.serializer_class().fields.items():
                res.append(
                    resources.Field(
                        attribute=field_name,
                        column_name=field.label,
                        readonly=True,
                    )
                )
        return res

    def export_resource(self, obj, selected_fields=None, **kwargs):
        if not selected_fields:
            selected_fields = self.get_export_fields()

        def _parse_representation(representation, template):
            template = re.sub(r"({{\W*}})", " ", template)
            t = Template(template).render(Context(representation))
            return t

        validated_data = []
        obj = dict(obj)
        for field in selected_fields:
            value = None
            if field.attribute in obj:
                value = obj[field.attribute]
                # Try to get the representation of this field derivated from the serializer field label key
                with suppress(KeyError):
                    representation_field_name = f"_{field.attribute}"
                    representation_field = self.serializer_class._declared_fields[representation_field_name]
                    if isinstance(representation_field, ListSerializer) or isinstance(
                        representation_field, RelatedListSerializer
                    ):
                        label_key = getattr(representation_field.child, "label_key", None)
                    else:
                        label_key = getattr(representation_field, "label_key", None)
                    if representation_field := obj.get(representation_field_name):
                        if isinstance(representation_field, list):
                            value = [
                                _parse_representation(representation, label_key)
                                for representation in representation_field
                            ]
                        else:
                            value = _parse_representation(representation_field, label_key)
            validated_data.append(value)
        return validated_data

    @classmethod
    def get_columns_map(cls, view):
        columns_map = {"id": "id"}
        # Get default fields from the list display
        if list_display := view.display_config_class(view, view.request, instance=None).get_list_display():
            if list_display.tree_group_field:
                columns_map[list_display.tree_group_field] = (
                    list_display.tree_group_label
                    if list_display.tree_group_label
                    else list_display.tree_group_field.title()
                )
            for key, label in list_display.flatten_fields:
                columns_map[key] = label
        return columns_map



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class DocumentModelRelationshipEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        if document_id := self.view.kwargs.get("document_id", None):
            return reverse(
                "wbcore:documents:document-documentmodelrelationship-list",
                args=[document_id],
                request=self.request,
            )
        return super().get_create_endpoint(**kwargs)
